import io
from contextlib import contextmanager
from copy import copy

import pytest
from alembic import op

import sqlalchemy as sa
from alembic.operations import Operations
from alembic.runtime.migration import MigrationContext

from keg.db import db
from keg_elements.db import migrations


class Context(MigrationContext):
    def statements(self):
        stmts = self.output_buffer.getvalue().splitlines(keepends=False)
        return [s for s in stmts if s]


class TestMigrationUtils:
    @contextmanager
    def get_context(self, dialect_name, connection=None, as_sql=True):
        dialect = getattr(sa.dialects, dialect_name).dialect()
        output = io.StringIO()
        context = Context(
            dialect,
            connection,
            dict(as_sql=as_sql, output_buffer=output, literal_binds=as_sql)
        )
        op._proxy = Operations(context)
        yield context
        del op._proxy

    def test_not_nullable_column_postgres(self):
        with self.get_context('postgresql') as context:
            migrations.add_non_nullable_column(
                op,
                'test_table',
                sa.Column('test_column', sa.Unicode),
                default_value='foo',
                schema='abc'
            )
        assert context.statements() == [
            'ALTER TABLE abc.test_table ADD COLUMN test_column VARCHAR;',
            'UPDATE abc.test_table SET test_column=\'foo\';',
            'ALTER TABLE abc.test_table ALTER COLUMN test_column SET NOT NULL;'
        ]

    def test_not_nullable_column_postgres_default_query(self):
        with self.get_context('postgresql') as context:
            column = sa.Column('test_column', sa.Unicode)

            migrations.add_non_nullable_column(
                op,
                'test_table',
                column,
                default_value_query=sa.table('test_table', copy(column)).update().values(
                    test_column=sa.func.random()),
            )
        assert context.statements() == [
            'ALTER TABLE test_table ADD COLUMN test_column VARCHAR;',
            'UPDATE test_table SET test_column=random();',
            'ALTER TABLE test_table ALTER COLUMN test_column SET NOT NULL;'
        ]

    def test_not_nullable_column_default_query_and_default(self):
        with self.get_context('postgresql'):
            column = sa.Column('test_column', sa.Unicode)

            with pytest.raises(ValueError,
                               match='You may not pass both default_value and default_value_query'):
                migrations.add_non_nullable_column(
                    op,
                    'test_table',
                    column,
                    default_value_query=sa.table('test_table', copy(column)).update().values(
                        test_column=sa.func.random()),
                    default_value='foo'
                )

    def test_not_nullable_column_no_default(self):
        with self.get_context('postgresql'):
            column = sa.Column('test_column', sa.Unicode)

            with pytest.raises(ValueError,
                               match='Must provide either default_value or default_value_query'):
                migrations.add_non_nullable_column(
                    op,
                    'test_table',
                    column
                )

    def test_not_nullable_column_mssql(self):
        with self.get_context('mssql') as context:
            migrations.add_non_nullable_column(
                op,
                'test_table',
                sa.Column('test_column', sa.Unicode),
                default_value='foo',
                schema='abc'
            )
        assert context.statements() == [
            'ALTER TABLE abc.test_table ADD test_column NVARCHAR(max) NULL;',
            'GO',
            'UPDATE abc.test_table SET test_column=N\'foo\';',
            'GO',
            'ALTER TABLE abc.test_table ALTER COLUMN test_column NVARCHAR(max) NOT NULL;',
            'GO'
        ]

    def test_not_nullable_column_sqlite(self):
        with self.get_context('sqlite') as context:
            migrations.add_non_nullable_column(
                op,
                'test_table',
                sa.Column('test_column', sa.Unicode),
                default_value='foo',
                schema='abc'
            )
        assert context.statements() == [
            'ALTER TABLE abc.test_table ADD COLUMN test_column VARCHAR;',
            'UPDATE abc.test_table SET test_column=\'foo\';',
            'ALTER TABLE abc.test_table ALTER COLUMN test_column SET NOT NULL;'
        ]

    def test_postgres_update_enum_options(self):
        with self.get_context('postgresql') as context:
            migrations.postgres_update_enum_options(
                op,
                [
                    ('public', 'test_table1', 'test_column1'),
                    ('abc', 'test_table2', 'test_column2'),
                ],
                'test_enum',
                ['value1', 'value2']
            )

        assert context.statements() == [
            'ALTER TABLE public.test_table1 ALTER COLUMN test_column1 TYPE VARCHAR USING test_column1::text;',  # noqa: E501
            'ALTER TABLE public.test_table1 ALTER COLUMN test_column1 DROP DEFAULT;',
            'ALTER TABLE abc.test_table2 ALTER COLUMN test_column2 TYPE VARCHAR USING test_column2::text;',  # noqa: E501
            'ALTER TABLE abc.test_table2 ALTER COLUMN test_column2 DROP DEFAULT;',
            'DROP TYPE test_enum;',
            "CREATE TYPE test_enum AS ENUM ('value1', 'value2');",
            'ALTER TABLE public.test_table1 ALTER COLUMN test_column1 TYPE test_enum USING test_column1::test_enum;',  # noqa: E501
            'ALTER TABLE abc.test_table2 ALTER COLUMN test_column2 TYPE test_enum USING test_column2::test_enum;',  # noqa: E501
        ]

    def test_postgres_get_enum_values(self):
        if db.engine.dialect.name != 'postgresql':
            pytest.skip('Postgres only test')

        with db.engine.begin() as conn:
            with self.get_context('postgresql', connection=conn, as_sql=False):
                sa.Enum('value1', 'value2', 'value3', name='test_enum').create(op.get_bind())

                results = migrations.postgres_get_enum_values(
                    op,
                    'test_enum',
                )
            assert results == ['value1', 'value2', 'value3']



from kegel_app.app import KegElApp


def cli_entry():
    KegElApp.cli.main()


@KegElApp.cli.command('verify-translations', help='Verifies all strings marked for translation')
def verify_translations():
    from pathlib import Path
    from morphi.messages.validation import check_translations

    root_path = Path(__file__).resolve().parent.parent
    check_translations(
        root_path,
        'keg_elements',
    )


if __name__ == '__main__':
    cli_entry()



import flask

from keg_elements.core import keg_element_blueprint
from keg_elements.views import FormView, GridView

from kegel_app import forms, grids


class DemoGrid(GridView):
    blueprint = keg_element_blueprint
    grid_cls = grids.DemoGrid


class DemoForm(FormView):
    blueprint = keg_element_blueprint
    form_cls = forms.DemoForm

    def form_on_valid(self):
        return flask.redirect('/demo-grid')



from . import SelectField

states = {
    'AK': 'Alaska',
    'AL': 'Alabama',
    'AR': 'Arkansas',
    'AS': 'American Samoa',
    'AZ': 'Arizona',
    'CA': 'California',
    'CO': 'Colorado',
    'CT': 'Connecticut',
    'DC': 'District of Columbia',
    'DE': 'Delaware',
    'FL': 'Florida',
    'GA': 'Georgia',
    'GU': 'Guam',
    'HI': 'Hawaii',
    'IA': 'Iowa',
    'ID': 'Idaho',
    'IL': 'Illinois',
    'IN': 'Indiana',
    'KS': 'Kansas',
    'KY': 'Kentucky',
    'LA': 'Louisiana',
    'MA': 'Massachusetts',
    'MD': 'Maryland',
    'ME': 'Maine',
    'MI': 'Michigan',
    'MN': 'Minnesota',
    'MO': 'Missouri',
    'MP': 'Northern Mariana Islands',
    'MS': 'Mississippi',
    'MT': 'Montana',
    'NA': 'National',
    'NC': 'North Carolina',
    'ND': 'North Dakota',
    'NE': 'Nebraska',
    'NH': 'New Hampshire',
    'NJ': 'New Jersey',
    'NM': 'New Mexico',
    'NV': 'Nevada',
    'NY': 'New York',
    'OH': 'Ohio',
    'OK': 'Oklahoma',
    'OR': 'Oregon',
    'PA': 'Pennsylvania',
    'PR': 'Puerto Rico',
    'RI': 'Rhode Island',
    'SC': 'South Carolina',
    'SD': 'South Dakota',
    'TN': 'Tennessee',
    'TX': 'Texas',
    'UT': 'Utah',
    'VA': 'Virginia',
    'VI': 'Virgin Islands',
    'VT': 'Vermont',
    'WA': 'Washington',
    'WI': 'Wisconsin',
    'WV': 'West Virginia',
    'WY': 'Wyoming',
}


class StateField(SelectField):
    """
    A U.S. States select field. Defaults to using short values (2 letter codes)
    for the stored value, and long labels for display.

    :param short_values: use 2 letter state codes for the stored value. Defaults to True.
    :param short_labels: use 2 letter state codes for the dropdown display. Defaults to False.
    """

    def __init__(self, *args, **kwargs):
        short_values = kwargs.pop('short_values', True)
        short_labels = kwargs.pop('short_labels', False)
        values = states.keys() if short_values else states.values()
        labels = states.keys() if short_labels else states.values()
        state_choices = list(zip(values, labels))
        super(StateField, self).__init__(choices=state_choices, *args, **kwargs)



import io
import os
import base64
from pathlib import Path

from keg_elements.encoding import force_bytes

import cryptography.fernet as fernet
import cryptography.hazmat.primitives.ciphers as ciphers
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import (
    constant_time,
    hashes,
    hmac,
    padding
)

from keg_elements.extensions import lazy_gettext as _


def aes_cipher(key, iv, mode=None):
    """Wrapper to build AES cipher from cryptography.

    :param key: AES key.
    :param iv: Used to create a CBC mode if a mode isn't manually provided.
    :param mode: Optional, provides an explicit mode to the cipher. CBC if not provided.
    :returns: Cipher instance.
    """
    algo = ciphers.algorithms.AES(key)
    mode = mode or ciphers.modes.CBC(iv)
    return ciphers.Cipher(algo, mode, backend=default_backend())


def aes_encryptor(key, mode=None):
    """Get encryptor cipher context for an AES key.

    Init vector is generated randomly.

    :param key: AES key.
    :param mode: Optional, provides an explicit mode to the cipher. CBC if not provided.
    :returns: Encryptor cipher context.
    """
    iv = os.urandom(ciphers.algorithms.AES.block_size // 8)
    cipher = aes_cipher(key, iv, mode=mode)
    return cipher.encryptor(), iv


def aes_decryptor(key, iv, mode=None):
    """Get decryptor cipher context for an AES key.

    :param key: AES key.
    :param iv: Initialization vector used to construct the default CBC mode.
    :param mode: Optional, provides an explicit mode to the cipher. CBC if not provided.
    :returns: Decryptor cipher context.
    """
    cipher = aes_cipher(key, iv=iv, mode=mode)
    return cipher.decryptor()


def fernet_cipher(key):
    """Build a Fernet cipher from the given key."""
    return fernet.Fernet(base64.urlsafe_b64encode(key))


def encrypt(data, key):
    """
    Encrypts binary data using cryptography's default fernet algorithm
    :param data: plaintext data to encrypt
    :param key: encryption key
    :return: encrypted data as a fernet token (a signed, base64 encoded string)
    """
    return fernet_cipher(key).encrypt(data)


def decrypt(data, key):
    """
    Decrypts binary data using cryptography's default fernet algorithm
    :param data: a fernet token to decrypt
    :param key: encryption key
    :return: decrypted data
    """
    return fernet_cipher(key).decrypt(data)


def encrypt_str(data, key):
    """
    Encrypts a unicode string using cryptography's default fernet algorithm
    :param data: unicode string to encrypt
    :param key: encryption key
    :return: encrypted utf-8 string as a fernet token
    """
    bin_data = data.encode('utf-8')
    return encrypt(bin_data, key)


def decrypt_str(cipher_text, key):
    """
    Decrypts a unicode string using cryptography's default fernet algorithm
    :param cipher_text: a utf-8 string as a fernet token to decrypt
    :param key: encryption key
    :return: decrypted unicode string
    """
    bin_data = decrypt(cipher_text, key)
    return bin_data.decode('utf-8')


def encrypt_file(key, in_fpath, out_fpath=None, chunksize=64 * 1024):
    """ Encrypts a file using AES (CBC mode) with the given key.

    :param key: The encryption key - a string that must be
        either 16, 24 or 32 bytes long. Longer keys
        are more secure.
    :param in_fpath: Full path of the input file
    :param out_fpath: Full path of the output file. If None, '<in_fpath>.enc' will be used.
    :param chunksize: Sets the size of the chunk which the function uses to read and encrypt
        the file. Larger chunk sizes can be faster for some files and machines. chunksize must
        be divisible by 16.
    :returns: Output file path string.
    """
    if not out_fpath:
        out_fpath = in_fpath + '.enc'

    with open(in_fpath, 'rb') as infile, open(out_fpath, 'wb') as outfile:
        for chunk in encrypt_fileobj(key, infile, chunksize):
            outfile.write(chunk)

    return out_fpath


def encrypt_fileobj(key, in_fileobj, chunksize=64 * 1024):
    """ Encrypts a file object using AES (CBC mode) with the given key.

    Example::

        with open('my_encrypted_file', mode='wb') as f:
            for chunk in encrypt_fileobj(my_crypto_key, buffer):
                f.write(chunk)

    :param key: The encryption key - a string that must be
        either 16, 24 or 32 bytes long. Longer keys
        are more secure.
    :param in_fpath: Full path of the input file
    :param out_fpath: Full path of the output file. If None, '<in_fpath>.enc' will be used.
    :param chunksize: Sets the size of the chunk which the function uses to read and encrypt
        the file. Larger chunk sizes can be faster for some files and machines. chunksize must
        be divisible by 16.
    :returns: Output file path string.
    """
    encryptor, iv = aes_encryptor(key)
    padder = padding.PKCS7(encryptor._ctx._cipher.block_size).padder()

    yield iv
    for chunk in iter(lambda: in_fileobj.read(chunksize), b''):
        data = padder.update(chunk)
        data = encryptor.update(data)
        yield data
    data = padder.finalize()
    yield encryptor.update(data)
    yield encryptor.finalize()


def decrypt_file(key, in_fpath, out_fpath=None, chunksize=24 * 1024):
    """ Decrypts a file using AES (CBC mode) with the
        given key. Parameters are similar to encrypt_file,
        with one difference: out_filename, if not supplied
        will be in_filename without its last extension
        (i.e. if in_filename is 'aaa.zip.enc' then
        out_filename will be 'aaa.zip')
    """
    if out_fpath is None:
        if Path(in_fpath).suffix == '.enc':
            out_fpath = Path(in_fpath).stem
        else:
            raise ValueError(_('If input file name doesn\'t end in ".enc" then output '
                               'filename must be given.'))
    with open(in_fpath, 'rb') as infile, open(out_fpath, 'wb') as outfile:
        decrypt_fileobj(key, infile, outfile, chunksize)

    return out_fpath


def decrypt_bytesio(key, in_fpath, chunksize=24 * 1024):
    """ Decrypts a file using AES (CBC mode) with the given key, and returns
    the contents in a BytesIO stream.

    :param key: The encryption key - a string that must be
        either 16, 24 or 32 bytes long. Longer keys
        are more secure.
    :param in_fpath: Full path of the input file
    :param chunksize: Sets the size of the chunk which the function uses to read and encrypt
        the file. Larger chunk sizes can be faster for some files and machines. chunksize must
        be divisible by 16.
    :returns: Output file path string.
    """
    with open(in_fpath, 'rb') as infile:
        bytes_fobj = io.BytesIO()
        decrypt_fileobj(key, infile, bytes_fobj, chunksize)
    # prep for reading from the start of the file
    bytes_fobj.seek(0)
    return bytes_fobj


def decrypt_fileobj(key, in_fileobj, out_fileobj, chunksize):
    """ Decrypts a file object using AES (CBC mode) with the given key, and writes
    the contents to the given output file object.

    :param key: The encryption key - a string that must be
        either 16, 24 or 32 bytes long. Longer keys
        are more secure.
    :param in_fileobj: Readable IO stream holding encrypted contents.
    :param out_fileobj: Writeable IO stream for decrypted contents.
    :param chunksize: Sets the size of the chunk which the function uses to read and encrypt
        the file. Larger chunk sizes can be faster for some files and machines. chunksize must
        be divisible by 16.
    :returns: None.
    """
    iv = in_fileobj.read(16)
    decryptor = aes_decryptor(key, iv)
    unpadder = padding.PKCS7(decryptor._ctx._cipher.block_size).unpadder()

    for chunk in iter(lambda: in_fileobj.read(chunksize), b''):
        data = decryptor.update(chunk)
        data = unpadder.update(data)
        out_fileobj.write(data)

    out_fileobj.write(unpadder.update(decryptor.finalize()))
    out_fileobj.write(unpadder.finalize())


def constant_time_compare(a, b):
    """Wrapper for cryptography constant time comparison, which will defeat timing attacks."""
    return constant_time.bytes_eq(a, b)


def salted_hmac(salt, value, secret):
    """Create an HMAC for a value using the given salt and secret.

    :param salt:
    :param value:
    :param secret:
    """
    salt = force_bytes(salt)
    secret = force_bytes(secret)
    value = force_bytes(value)

    hasher = hashes.Hash(hashes.SHA512(), backend=default_backend())
    hasher.update(salt + secret)
    key = hasher.finalize()

    hmacer = hmac.HMAC(key, hashes.SHA512(), backend=default_backend())
    hmacer.update(value)
    return hmacer.finalize()



import flask
from keg.web import BaseView

from keg_elements.extensions import gettext as _


class FormMixin:
    """Mixin supporting simple form view setup"""
    form_cls = None
    """form class to construct, or callable returning a form instance."""
    flash_success = ('Form submitted successfully.', 'success')
    """flash message to show on post when form data has passed validation"""
    flash_failure = ('Form errors detected, see below for details.', 'error')
    """flash message to show on post when form data has failed validation"""
    title = None
    """Page title, will be assigned as title for the template."""

    def form_default(self):
        """Returns default object to load in form during init."""
        return

    def form_create(self):
        """Create the form instance with default object (if any)."""
        default = self.form_default()
        return self.form_cls(obj=default)

    def form_init(self):
        """Assign a form instance to the template."""
        self.form = self.form_create()
        self.assign('form', self.form)

    def get(self):
        """GET method responder. By default, inits form to template."""
        self.assign('title', self.title)
        self.form_init()

    def post(self):
        """POST method responder.

        Initializes the form and attempts to validate POSTed data. Flashes messages
        if those attributes are set, then directs to on_form_[in]valid
        as needed for additional workflow.
        """
        self.form_init()
        if self.form.validate():
            if self.flash_success:
                flask.flash(*self.flash_success)
            self.assign('form_success', True)
            return self.form_on_valid()
        else:
            if self.flash_failure:
                flask.flash(*self.flash_failure)
            return self.form_on_invalid()

    def form_on_valid(self):
        """Action method for form that has passed validation.

        This is usually where things like database transactions will go.
        """
        pass

    def form_on_invalid(self):
        """Action method for form that has failed validation."""
        pass


class FormView(FormMixin, BaseView):
    """Base view for simple form pages."""


class GridMixin:
    """Mixin supporting simple grid view setup"""
    grid_cls = None
    """Grid class to construct, or callable returning a grid instance."""
    template = 'keg-elements/grid-view.html'
    """Template to render. Defaults to keg-elements/grid-view.html"""
    title = None
    """Page title, will be assigned as title for the template."""

    def get(self):
        """GET method responder. By default, fully processes and renders grid to template."""
        return self.render_grid()

    def allow_post(self, grid):
        """Inspects the grid manager's arg loads to determine if post should be allowed."""
        from webgrid.extensions import RequestFormLoader
        if grid.manager and RequestFormLoader not in grid.manager.args_loaders:
            flask.abort(405)

    def post(self):
        """POST method responder.

        For session-enabled grids (the default), process the args, which will save any form
        args into the grid's session store. Then, redirect to a GET request on the same URL.

        If the grid is not session-enabled, render to the template.

        Note: for request form args to be processed, the grid's manager must be set up with
        an args loader that uses them.
        """
        if not self.grid_cls.session_on:
            return self.render_grid()
        self.allow_post(
            self.process_grid()
        )
        return flask.redirect(flask.request.full_path)

    def init_grid(self):
        """ Create an instance of `grid_cls` """
        return self.grid_cls()

    def process_grid(self):
        """Create grid instance and run args collection/processing for grid state."""
        if self.grid_cls is None:
            raise NotImplementedError(
                'You must set {}.grid_cls to render a grid'.format(self.__class__.__name__)
            )

        grid = self.init_grid()
        grid.apply_qs_args()

        return self.post_args_grid_setup(grid)

    def post_args_grid_setup(self, grid):
        """ Apply changes to grid instance after QS args/session are loaded """
        return grid

    def get_grid_template_args(self, arg_dict):
        """ Return any template arguments needed for render.

        By default, `grid` and `title` arguments are passed in.
        """
        return arg_dict

    def on_render_limit_exceeded(self, grid):
        """ Handles export limit exceeded case. By default, shows a flash message."""
        flask.flash(_('Too many records to export as {}').format(grid.export_to), 'error')

    def render_grid(self):
        """ Build grid, handle loading args and conditional export, and render result."""
        grid = self.process_grid()

        if grid.export_to:
            import webgrid

            try:
                return grid.export_as_response()
            except webgrid.renderers.RenderLimitExceeded:
                self.on_render_limit_exceeded(grid)

        # args added with self.assign should be passed through here
        template_args = self.get_grid_template_args(dict(self.template_args, **{
            'grid': grid,
            'title': self.title,
        }))

        return flask.render_template(self.template, **template_args)


class GridView(GridMixin, BaseView):
    """Base view for simple grid pages."""



import datetime
from decimal import Decimal

from keg.db import db
import arrow
import freezegun
import pytest
import sqlalchemy as sa
import sqlalchemy_utils as sautils
import keg_elements.db.mixins as mixins
import keg_elements.db.columns as columns

import kegel_app.model.entities as ents


class TestDefaultColsMixin:

    def setup_method(self, fn):
        ents.Thing.delete_cascaded()

    def test_default_ordering(self):
        if db.engine.dialect.name != 'sqlite':
            pytest.skip('SQLite only test')

        thing1 = ents.Thing.fake(id=6)
        thing2 = ents.Thing.fake(id=5)
        thing3 = ents.Thing.fake(id=7)

        assert ents.Thing.query.all() == [thing2, thing1, thing3]

    def test_utc_default(self):
        with freezegun.freeze_time('2020-01-01 13:01:01', tz_offset=4):
            thing = ents.Thing.fake()
        assert thing.created_utc.format() == '2020-01-01 17:01:01+00:00'
        assert thing.updated_utc.format() == '2020-01-01 17:01:01+00:00'

        with freezegun.freeze_time('2020-02-01 14:03:05', tz_offset=3):
            db.session.query(ents.Thing).filter_by(id=thing.id).update({})
            db.session.commit()
        assert thing.updated_utc.format() == '2020-02-01 17:03:05+00:00'


class TestMethodsMixin:

    def setup_method(self, fn):
        ents.Thing.delete_cascaded()
        ents.MultiplePrimaryKeys.delete_cascaded()

    def test_add(self):
        ents.Thing.add(name='name', color='color', scale_check=1)
        assert ents.Thing.query.count() == 1

        row = ents.Thing.query.first()

        assert row.name == 'name'
        assert row.color == 'color'
        assert row.scale_check == 1

    def test_insert(self):
        ret_id = ents.Thing.insert(name='name', color='color', scale_check=1)
        assert ents.Thing.query.count() == 1

        row = ents.Thing.query.first()

        if db.engine.dialect.name != 'sqlite':
            assert row.id == ret_id
        assert row.name == 'name'
        assert row.color == 'color'
        assert row.scale_check == 1

    def test_insert_values_dict(self):
        ents.Thing.insert(values={'name': 'name'}, color='color', scale_check=1)
        row = ents.Thing.query.first()

        assert row.name == 'name'
        assert row.color == 'color'
        assert row.scale_check == 1

    def test_insert_multiple_pk(self):
        ret_id = ents.MultiplePrimaryKeys.insert(name='name', id=54, other_pk=5)
        assert ents.MultiplePrimaryKeys.query.count() == 1

        row = ents.MultiplePrimaryKeys.query.first()

        if db.engine.dialect.name != 'sqlite':
            id_key = (54, 5) if ents.MultiplePrimaryKeys.primary_keys()[0].name == 'id' else (5, 54)
            assert id_key == ret_id
        assert row.name == 'name'

    def test_delete(self):
        thing = ents.Thing.fake()
        assert ents.Thing.query.count() == 1

        ents.Thing.delete(thing.id)
        assert ents.Thing.query.count() == 0

        assert ents.Thing.get(1) is None
        assert ents.Thing.delete(1) is False

    def test_delete_cascaded(self):
        ents.Thing.fake()
        assert ents.Thing.query.count() == 1

        ents.Thing.delete_cascaded()
        assert ents.Thing.query.count() == 0

    def test_edit(self):
        thing1 = ents.Thing.fake()

        ents.Thing.edit(_oid=thing1.id, name='edited')
        assert thing1.name == 'edited'

        with pytest.raises(AttributeError):
            ents.Thing.edit(name='edited')

    def test_update(self):
        thing1 = ents.Thing.fake()

        ents.Thing.update(thing1.id, name='edited', color='silver')
        db.session.commit()
        db.session.refresh(thing1)

        assert thing1.name == 'edited'
        assert thing1.color == 'silver'

    def test_update_values_dict(self):
        thing1 = ents.Thing.fake()

        ents.Thing.update(thing1.id, values={'name': 'edited'}, color='silver')
        db.session.commit()
        db.session.refresh(thing1)

        assert thing1.name == 'edited'
        assert thing1.color == 'silver'

    def test_update_multiple_pk(self):
        row = ents.MultiplePrimaryKeys.fake(id=55, other_pk=6, name='foo')
        id_key = (55, 6) if ents.MultiplePrimaryKeys.primary_keys()[0].name == 'id' else (6, 55)
        ents.MultiplePrimaryKeys.update(id_key, name='bar')
        db.session.commit()
        db.session.refresh(row)

        assert row.name == 'bar'

    def test_from_dict(self):
        # Testing create uses `from_dict` so create objects by hand
        related = ents.RelatedThing(name='something')
        db.session.add(related)

        related.from_dict({
            'key_not_present': 'blah'
        })
        assert related.name == 'something'

        related.from_dict({'name': 'blah'})
        assert related.name == 'blah'

        related.from_dict({'thing': {'name': 'thing_name'}})
        assert related.thing.name == 'thing_name'

        related.thing.from_dict({'manys': [{'name': 'many_1'}, {'name': 'many_2'}]})
        assert ents.ManyToManyThing.query.count() == 2
        assert len(related.thing.manys) == 2

    def test_add_edit(self):
        obj = ents.MultiplePrimaryKeys.add_or_edit({
            'name': 'something', 'id': 1, 'other_pk': 1})

        assert obj.name == 'something'
        assert obj.id == 1
        assert obj.other_pk == 1

        with pytest.raises(sa.exc.InvalidRequestError, match='Incorrect number of values'):
            ents.MultiplePrimaryKeys.add_or_edit({'name': 'other', 'id': 1})

        with pytest.raises(sa.exc.InvalidRequestError, match='Incorrect number of values'):
            ents.MultiplePrimaryKeys.add_or_edit({'name': 'other', 'other_pk': 1})

        ents.MultiplePrimaryKeys.add_or_edit({'name': 'other', 'id': 1, 'other_pk': 1})
        assert obj.name == 'other'

    def test_get(self):
        thing = ents.Thing.fake(name='banana')
        assert ents.Thing.get(thing.id).name == 'banana'
        ents.MultiplePrimaryKeys.fake(id=1, other_pk=2, name='banana')
        id_key = (1, 2) if ents.MultiplePrimaryKeys.primary_keys()[0].name == 'id' else (2, 1)
        assert ents.MultiplePrimaryKeys.get(id_key).name == 'banana'
        assert ents.MultiplePrimaryKeys.get({'id': 1, 'other_pk': 2}).name == 'banana'

    def test_get_by(self):

        def returns_none_when_count_is_zero():
            assert ents.Thing.get_by(id=1) is None

        def returns_object_when_one_exists():
            thing = ents.Thing.fake()

            assert ents.Thing.get_by(id=thing.id) == thing

        def raises_when_multiple_exists():
            ents.Thing.fake(name='foo')
            ents.Thing.fake(name='foo')

            with pytest.raises(sa.orm.exc.MultipleResultsFound):
                ents.Thing.get_by(name='foo')

        returns_none_when_count_is_zero()
        returns_object_when_one_exists()
        raises_when_multiple_exists()

    def test_get_where(self):
        thing1 = ents.Thing.fake(name='thing1', color='black')
        ents.Thing.fake(name='thing2', color='black', scale_check=10)
        thing3 = ents.Thing.fake(name='thing3', color='orange', scale_check=10)

        assert ents.Thing.get_where(ents.Thing.name == 'thing1') == thing1
        assert ents.Thing.get_where(ents.Thing.name == 'thing5') is None
        assert (ents.Thing.get_where(ents.Thing.color == 'orange',
                                     ents.Thing.scale_check == 10) == thing3)

        with pytest.raises(sa.orm.exc.MultipleResultsFound):
            ents.Thing.get_where(ents.Thing.color == 'black')

    def test_pairs_takes_name_and_value_and_returns_list_of_tuples(self):
        thing1 = ents.Thing.fake(name='A', color='orange')
        thing2 = ents.Thing.fake(name='B', color='black')

        expected = [(thing1.id, 'A'), (thing2.id, 'B')]
        assert ents.Thing.pairs('id', 'name') == expected

        expected = [(thing2.id, 'B'), (thing1.id, 'A')]
        assert ents.Thing.pairs('id', 'name',
                                order_by=[ents.Thing.name.desc()]) == expected

        expected = [(thing2.id, 'B')]
        query = ents.Thing.query.filter_by(name='B')
        result = ents.Thing.pairs('id', 'name', query=query)
        assert result == expected

        expected = [thing1.id, thing2.id]
        assert ents.Thing.pairs('id', 'name',
                                items=lambda obj: obj.id) == expected

    def test_to_dict(self):
        obj = ents.Thing.fake()

        expected = {'id', 'name', 'color', 'scale_check', 'float_check', 'updated_utc',
                    'created_utc', 'units', 'status', 'date_check', 'float_check_prop'}
        assert set(obj.to_dict().keys()) == expected
        assert set(obj.to_dict(exclude={'id'}).keys()) == expected - {'id'}

        hybrids = ['name_and_color']
        exclude = ['name', 'scale_check', 'float_check', 'color', 'updated_utc', 'created_utc',
                   'units', 'status', 'date_check']
        dictionary = obj.to_dict(hybrids=hybrids, exclude=exclude)

        assert dictionary == {
            'id': obj.id,
            'name_and_color': obj.name_and_color,
            'float_check_prop': obj.float_check_prop,
        }

    def test_random_data_for_column(self):
        func = mixins.MethodsMixin.random_data_for_column

        assert type(func(sa.Column(sa.Unicode), None)) == str
        assert type(func(sa.Column(sa.String), None)) == str
        assert type(func(sa.Column(sa.Integer), None)) == int
        assert type(func(sa.Column(sa.Boolean), None)) == bool
        assert type(func(sa.Column(sa.Numeric(1, 1)), None)) == Decimal
        assert type(func(sa.Column(sa.Numeric), None)) == float
        assert type(func(sa.Column(sa.Float), None)) == float
        assert type(func(sa.Column(sa.Date), None)) == datetime.date
        assert type(func(sa.Column(sa.DateTime), None)) == datetime.datetime
        assert type(func(sa.Column(sautils.ArrowType), None)) == arrow.Arrow
        assert isinstance(func(sa.Column(sautils.EmailType), None), str)
        assert isinstance(func(sa.Column(columns.TimeZoneType), None), str)

        with pytest.raises(ValueError):
            assert type(func(sa.Column(sa.LargeBinary), None))

    def test_check_kwargs_in_fake(self):
        ents.Thing.fake(name='a')

        with pytest.raises(AssertionError) as excinfo:
            ents.Thing.fake(foo=1, bar=2, _baz=3, name='b')
        assert str(excinfo.value) == \
            'Unknown column or relationship names in kwargs: [\'bar\', \'foo\']'

        ents.Thing.fake(foo=1, bar=2, _baz=3, name='b', _check_kwargs=False)

    def test_check_kwargs_in_edit(self):
        thing = ents.Thing.fake(name='a')
        with pytest.raises(AssertionError) as excinfo:
            ents.Thing.edit(thing.id, fieldshouldnotexist='foo')
        assert str(excinfo.value) == \
            'Unknown column or relationship names in kwargs: [\'fieldshouldnotexist\']'

        ents.Thing.edit(thing.id, fieldshouldnotexist='foo', _check_kwargs=False)

    def test_check_kwargs_in_add(self):
        with pytest.raises(AssertionError) as excinfo:
            ents.Thing.add(name='kwargs-in-add', fieldshouldnotexist='foo')
        assert str(excinfo.value) == \
            'Unknown column or relationship names in kwargs: [\'fieldshouldnotexist\']'

        db.session.remove()
        ents.Thing.add(name='kwargs-in-add', fieldshouldnotexist='foo', _check_kwargs=False)

    def test_testing_create_flush_and_commit(self):
        obj = ents.Thing.fake(_flush=False, _commit=False)
        assert sa.inspect(obj).pending
        db.session.rollback()
        assert sa.inspect(obj).transient

        obj = ents.Thing.fake(_flush=True, _commit=False)
        assert sa.inspect(obj).persistent
        db.session.rollback()
        assert sa.inspect(obj).transient

        obj = ents.Thing.fake(_flush=False, _commit=True)
        assert sa.inspect(obj).persistent
        db.session.rollback()
        assert sa.inspect(obj).persistent

    def test_override_random_data_generation(self):
        obj = ents.Thing.fake()
        assert obj.color == 'blue'
        assert obj.scale_check == Decimal('12.3456')

    def test_related_object_created(self):
        obj = ents.RelatedThing.fake()
        assert obj.thing

    def test_related_object_respects_id(self):
        thing = ents.Thing.fake()
        obj = ents.RelatedThing.fake(thing_id=thing.id)
        assert obj.thing is thing

    def test_related_object_respects_relationship_attr(self):
        thing = ents.Thing.fake()
        obj = ents.RelatedThing.fake(thing=thing)
        assert obj.thing is thing

    def test_related_object_bad_attributes(self):
        class MyObject(mixins.MethodsMixin):
            pass

        with pytest.raises(Exception, match='expects "thing" on class MyObject'):
            MyObject.testing_set_related({}, ents.Thing)

    def test_related_object_takes_overrides(self):
        class MyObject(mixins.MethodsMixin):
            foo = None
            bar = None

        kwargs = {}
        MyObject.testing_set_related(
            kwargs, ents.Thing, _relationship_name='foo', _relationship_field='bar'
        )
        assert isinstance(kwargs['foo'], ents.Thing)

        thing = ents.Thing.fake()
        kwargs = {'foo': thing}
        MyObject.testing_set_related(
            kwargs, ents.Thing, _relationship_name='foo', _relationship_field='bar'
        )
        assert kwargs['foo'] is thing

        kwargs = {'bar': thing.id}
        MyObject.testing_set_related(
            kwargs, ents.Thing, _relationship_name='foo', _relationship_field='bar'
        )
        assert kwargs['bar'] == thing.id
        assert 'foo' not in kwargs


class TestSoftDeleteMixin:

    def setup_method(self, _):
        ents.SoftDeleteTester.delete_cascaded()
        ents.HardDeleteParent.delete_cascaded()
        db.session.remove()

    def test_delete_sets_field(self):
        sdt1 = ents.SoftDeleteTester.fake()

        assert sdt1.deleted_utc is None
        assert ents.SoftDeleteTester.delete(sdt1.id)
        assert sdt1.deleted_utc is not None

    def test_manual_delete_raises_error(self):
        sdt1 = ents.SoftDeleteTester.fake()

        db.session.delete(sdt1)

        with pytest.raises(mixins.HardDeleteProhibitedError):
            db.session.commit()

        db.session.rollback()

    def test_delete_without_id_returns_none(self):
        assert not ents.SoftDeleteTester.delete(1234)

    def test_delete_cascaded_still_works(self):
        ents.SoftDeleteTester.fake()

        assert ents.SoftDeleteTester.query.count() == 1

        ents.SoftDeleteTester.delete_cascaded()
        assert ents.SoftDeleteTester.query.count() == 0

    def test_testing_create_allows_is_deleted_flag(self):
        sdt1 = ents.SoftDeleteTester.fake()
        assert sdt1.deleted_utc is None

        sdt1 = ents.SoftDeleteTester.fake(_is_deleted=True)
        assert sdt1.deleted_utc is not None

    def test_deleting_the_parent_deletes_the_child(self):
        sdt1 = ents.SoftDeleteTester.fake()
        assert sdt1.hdp == ents.HardDeleteParent.query.one()

        with pytest.raises(mixins.HardDeleteProhibitedError):
            sdt1.hdp.delete(sdt1.hdp_id)


class TestLookupMixin:
    def setup_method(self):
        ents.LookupTester.delete_cascaded()

    def _labels_in_list(labels, rows):
        row_labels = [row.label for row in rows]
        for label in labels:
            if label not in row_labels:
                return False

        return True

    def test_list_active(self):
        a = ents.LookupTester.fake(label='a', deleted_utc=None)
        b = ents.LookupTester.fake(label='b', deleted_utc=None)
        c = ents.LookupTester.fake(label='c', deleted_utc=None)
        d = ents.LookupTester.fake(label='d', deleted_utc=arrow.now())
        e = ents.LookupTester.fake(label='e', deleted_utc=arrow.now())

        rows = ents.LookupTester.list_active()
        assert [a, b, c] == rows

        rows = ents.LookupTester.list_active(order_by=ents.LookupTester.label.desc())
        assert [c, b, a] == rows

        rows = ents.LookupTester.list_active(include_ids=d.id)
        assert [a, b, c, d] == rows

        rows = ents.LookupTester.list_active(include_ids=(d.id, e.id))
        assert [a, b, c, d, e] == rows

    def test_pairs_active(self):
        a = ents.LookupTester.fake(label='a', deleted_utc=None)
        b = ents.LookupTester.fake(label='b', deleted_utc=None)
        c = ents.LookupTester.fake(label='c', deleted_utc=None)
        d = ents.LookupTester.fake(label='d', deleted_utc=arrow.now())
        e = ents.LookupTester.fake(label='e', deleted_utc=arrow.now())

        def make_pairs(*records):
            return [(record.id, record.label) for record in records]

        pairs = ents.LookupTester.pairs_active()
        assert make_pairs(a, b, c) == pairs

        pairs = ents.LookupTester.pairs_active(order_by=ents.LookupTester.label.desc())
        assert make_pairs(c, b, a) == pairs

        pairs = ents.LookupTester.pairs_active(include_ids=d.id)
        assert make_pairs(a, b, c, d) == pairs

        pairs = ents.LookupTester.pairs_active(include_ids=(d.id, e.id))
        assert make_pairs(a, b, c, d, e) == pairs

    def test_get_by_label(self):
        a = ents.LookupTester.fake(label='a', deleted_utc=None)
        b = ents.LookupTester.fake(label='b', deleted_utc=None)

        assert ents.LookupTester.get_by_label('a') == a
        assert ents.LookupTester.get_by_label('b') == b

    def test_get_by_code(self):
        a = ents.LookupTester.fake(label='a', code='foo', deleted_utc=None)
        ents.LookupTester.fake(label='b', deleted_utc=None)

        assert ents.LookupTester.get_by_code('foo') == a

    def test_repr(self):
        a = ents.LookupTester.fake(label='a', deleted_utc=None)
        assert str(a) == '<LookupTester {}:a>'.format(a.id)



"""Common SQLAlchemy column types."""
import enum
import random

import sqlalchemy as sa

from keg_elements import crypto
from keg_elements.extensions import lazy_gettext as _


class TimeZoneType(sa.Unicode):
    """A column type for time zones, stored as a unicode string."""

    def __init__(self, length=None, **kwargs):  # Override any given length.
        super(TimeZoneType, self).__init__(length=255, **kwargs)


class EncryptedUnicode(sa.TypeDecorator):
    """
    Unicode column type that encrypts value with the given key for persistance to storage.

    :param key: A bytes object containing the encryption key or a callable that returns the key
    :param encrypt: A callable that takes a unicode string and the encryption key as arguments
        and returns the encrypted data as a bytes object.
    :param decrypt: A callable that takes a bytes object and the encryption key as arguments
        and returns the decrypted data as a unicode string.
    """
    impl = sa.UnicodeText

    def __init__(self, *args, **kwargs):
        self._key = kwargs.pop('key')
        self._encrypt = kwargs.pop('encrypt', crypto.encrypt_str)
        self._decrypt = kwargs.pop('decrypt', crypto.decrypt_str)
        super(EncryptedUnicode, self).__init__(*args, **kwargs)

    @property
    def key(self):
        key_val = self._key() if callable(self._key) else self._key
        if len(key_val) < 32:
            raise ValueError(_('Key must be at least 32 bytes long'))
        return key_val[:32]

    def process_bind_param(self, value, dialect):
        if value is None:
            return None
        return self._encrypt(value, self.key).decode()

    def process_result_value(self, value, dialect):
        if value is None:
            return None
        return self._decrypt(value.encode(), self.key)


class DBEnum(enum.Enum):
    """
    Base class for all database enum types.

    To create a new enum, subclass this, add the enum values, and implement db_name()::

        class MyEnum(DBEnum):
            option1 = 'Option 1'
            option2 = 'Option 2'

            @classmethod
            def db_name(cls):
                return 'my_enum_db_name'

    To declare a DB column of this type::

        class MyEntity(db.Model):
            option = sa.Column(MyEnum.db_type())

    To set the choices on a form field::

        class MyEntityForm(wtforms.Form):
            option = wtforms.SelectField(
                'Option',
                choices=MyEnum.form_options()
                coerce=MyEnum.coerce
            )

    If using `ModelForm` the field will be configured with the above options automatically::

        class MyEntityForm(ModelForm):
            class Meta:
                model = MyEntity

    """
    @classmethod
    def db_name(cls):
        raise NotImplementedError

    @classmethod
    def db_type(cls):
        return sa.Enum(cls, name=cls.db_name())

    @classmethod
    def option_pairs(cls):
        return [(o.name, o.value) for o in cls]

    @classmethod
    def form_options(cls):
        return [(o, o.value) for o in cls]

    @classmethod
    def coerce(cls, value):
        if value is None:
            return None
        if isinstance(value, cls):
            return value
        try:
            return cls[value]
        except KeyError:
            raise ValueError('Not a valid selection')

    @classmethod
    def random(cls):
        return random.choice(list(cls))

    def __str__(self):
        return self.name

    def __json__(self):
        return str(self)



from unittest import mock

import flask
import pytest

from keg_elements import sentry


class TestSentryFilter:
    def test_repr_process(self):
        f = sentry.SentryEventFilter()

        conf = flask.config.Config(__file__)
        assert f.repr_process(conf, {}) == '<Filtered Config>'
        assert f.repr_process(dict(), {}) is NotImplemented
        assert f.repr_process(1, {}) is NotImplemented
        assert f.repr_process('', {}) is NotImplemented

    def test_filter_by_key(self):
        event = {
            'exception': {
                'stacktrace': {
                    'frames': [
                        {
                            'module': 'foo.bar',
                            'vars': {
                                'secret_key': '1234',
                                'safe': 'abc'
                            }
                        },
                        {
                            'module': 'foo.baz',
                            'vars': {
                                'password': '4321',
                                'unchanged': 'def'
                            }
                        },

                    ]
                },
                'token': 1
            }
        }
        filtered = sentry.SentryEventFilter().before_send(event, {})
        assert filtered == {
            'exception': {
                'stacktrace': {
                    'frames': [
                        {
                            'module': 'foo.bar',
                            'vars': {
                                'secret_key': '<Filtered str>',
                                'safe': 'abc'
                            }
                        },
                        {
                            'module': 'foo.baz',
                            'vars': {
                                'password': '<Filtered str>',
                                'unchanged': 'def'
                            }
                        },

                    ]
                },
                'token': '<Filtered int>'
            }
        }

    def test_filter_by_module(self):
        event = {
            'exception': {
                'stacktrace': {
                    'frames': [
                        {
                            'module': 'foo.app',
                            'vars': {
                                'unchanged': 'xyz',
                            }
                        },
                        {
                            'module': 'cryptography',
                            'vars': {
                                'secret_key': '1234',
                                'safe': 'abc'
                            }
                        },
                        {
                            'module': 'passlib.submodule',
                            'vars': {
                                'password': '4321',
                                'unchanged': 'def'
                            }
                        },
                        {
                            'module': 'foo.libs',
                            'vars': {
                                'password': '4321',
                                'unchanged': 'def'
                            }
                        },
                    ]
                },
            }
        }
        filtered = sentry.SentryEventFilter().before_send(event, {})
        assert filtered == {
            'exception': {
                'stacktrace': {
                    'frames': [
                        {
                            'module': 'foo.app',
                            'vars': {
                                'unchanged': 'xyz'
                            }
                        },
                        {
                            'module': 'cryptography',
                            'vars': {}
                        },
                        {
                            'module': 'passlib.submodule',
                            'vars': {}
                        },
                        {
                            'module': 'foo.libs',
                            'vars': {}
                        },

                    ]
                },
            }
        }

    def test_filter_modules_no_stacktrace(self):
        event = {
            'exception': {
                'values': [
                    {
                        'foo': 'bar'
                    }
                ]
            }
        }
        filtered = sentry.SentryEventFilter().before_send(event, {})
        assert filtered == {
            'exception': {
                'values': [
                    {
                        'foo': 'bar',
                        'stacktrace': {
                            'frames': []
                        }
                    }
                ]
            }
        }

    def test_filter_by_module_multiple_exceptions(self):
        event = {
            'exception': {
                'values': [
                    {
                        'stacktrace': {
                            'frames': [
                                {
                                    'module': 'foo.app',
                                    'vars': {
                                        'unchanged': 'xyz',
                                    }
                                },
                                {
                                    'module': 'cryptography',
                                    'vars': {
                                        'safe': 'abc'
                                    }
                                },
                            ]
                        },
                    },
                    {
                        'stacktrace': {
                            'frames': [
                                {
                                    'module': 'foo.bar',
                                    'vars': {
                                        'unchanged': 'abc',
                                    }
                                },
                                {
                                    'module': 'keg_elements.crypto.encrypt',
                                    'vars': {
                                        'safe': 'abc'
                                    }
                                },
                            ]
                        },
                    },

                ]
            }
        }
        filtered = sentry.SentryEventFilter().before_send(event, {})
        assert filtered == {
            'exception': {
                'values': [
                    {
                        'stacktrace': {
                            'frames': [
                                {
                                    'module': 'foo.app',
                                    'vars': {
                                        'unchanged': 'xyz'
                                    }
                                },
                                {
                                    'module': 'cryptography',
                                    'vars': {}
                                },
                            ]
                        }
                    },
                    {
                        'stacktrace': {
                            'frames': [
                                {
                                    'module': 'foo.bar',
                                    'vars': {
                                        'unchanged': 'abc',
                                    }
                                },
                                {
                                    'module': 'keg_elements.crypto.encrypt',
                                    'vars': {}
                                },
                            ]
                        },
                    },
                ]

            }
        }

    def test_filter_request(self):
        event = {
            'request': {
                'cookies': {
                    '_xsrf': 'foo',
                    'session': 'ol9auwkerjevt1bzvopg9vri311o5snc.naif3fmfext6l2shp6kr78zkwn'
                },
                'env': {
                    'REMOTE_ADDR': '127.0.0.1',
                    'SERVER_NAME': '127.0.0.1',
                    'SERVER_PORT': '5000'
                },
                'headers': {
                    'Accept-Encoding': 'gzip, deflate, br',
                    'Accept-Language': 'en-US,en;q=0.9',
                    'Cache-Control': 'max-age=0',
                    'Connection': 'keep-alive',
                    'Cookie': 'session=ol9auwkerjevt1bzvopg9vri311o5snc.naif3fmfext6l2shp6kr78zkwn',
                    'Dnt': '1',
                    'Host': 'localhost:5000',
                    'Upgrade-Insecure-Requests': '1',
                },
                'method': 'GET',
                'query_string': '',
                'url': 'http://localhost:5000/error'
            }
        }
        filtered = sentry.SentryEventFilter().before_send(event, {})
        assert filtered == {
            'request': {
                'cookies': {
                    '_xsrf': '<Filtered str>',
                    'session': '<Filtered str>'
                },
                'env': {
                    'REMOTE_ADDR': '127.0.0.1',
                    'SERVER_NAME': '127.0.0.1',
                    'SERVER_PORT': '5000'
                },
                'headers': {
                    'Accept-Encoding': 'gzip, deflate, br',
                    'Accept-Language': 'en-US,en;q=0.9',
                    'Cache-Control': 'max-age=0',
                    'Connection': 'keep-alive',
                    'Cookie': '<Filtered str>',
                    'Dnt': '1',
                    'Host': 'localhost:5000',
                    'Upgrade-Insecure-Requests': '1',
                },
                'method': 'GET',
                'query_string': '',
                'url': 'http://localhost:5000/error'
            }
        }

    def test_filter_by_config_value(self, monkeypatch):
        monkeypatch.setitem(flask.current_app.config, 'STRING', '1234567890')
        monkeypatch.setitem(flask.current_app.config, 'BYTES1', b'\xF0\x0B\xAA')
        monkeypatch.setitem(flask.current_app.config, 'BYTES2', b'hello world')
        monkeypatch.setitem(flask.current_app.config, 'IGNORED', 'ignored')
        monkeypatch.setitem(flask.current_app.config, 'EMPTY', '')
        monkeypatch.setitem(flask.current_app.config, 'NONE', None)

        class Filter(sentry.SentryEventFilter):
            sanitized_config_keys = [
                'STRING',
                'BYTES1',
                'BYTES2',
                'EMPTY',
                'NONE',
                'UNSET',
            ]

        event = {
            'exception': {
                'stacktrace': {
                    'frames': [
                        {
                            'module': 'foo.bar',
                            'vars': {
                                'string1': '**1234567890**',
                                'string2': '__31323334353637383930__',
                                'string3': '**MTIzNDU2Nzg5MA==**',
                                'string4': '**MTIzNDU2Nzg5MA**',
                                'bytes1_1': '**F00BAA**',
                                'bytes1_2': '**8Auq**',
                                'bytes1_3': '**%F0%0B%AA**',
                                'bytes2_1': '**hello world**',
                                'bytes2_2': '**hello+world**',
                                'bytes2_3': '**aGVsbG8gd29ybGQ=**',
                                'multiple': '**1234567890**hello world**',
                                'ignored': '**ignored**',
                                'url': 'http://localhost?var1=hello+world&var2=%F0%0B%AA',
                                'empty': '',
                                'none': None,
                            }
                        },
                    ]
                },
            }
        }
        filtered = Filter().before_send(event, {})
        assert filtered == {
            'exception': {
                'stacktrace': {
                    'frames': [
                        {
                            'module': 'foo.bar',
                            'vars': {
                                'string1': '**<Filtered str>**',
                                'string2': '__<Filtered str>__',
                                'string3': '**<Filtered str>**',
                                'string4': '**<Filtered str>**',
                                'bytes1_1': '**<Filtered str>**',
                                'bytes1_2': '**<Filtered str>**',
                                'bytes1_3': '**<Filtered str>**',
                                'bytes2_1': '**<Filtered str>**',
                                'bytes2_2': '**<Filtered str>**',
                                'bytes2_3': '**<Filtered str>**',
                                'multiple': '**<Filtered str>**<Filtered str>**',
                                'ignored': '**ignored**',
                                'url': 'http://localhost?var1=<Filtered str>&var2=<Filtered str>',
                                'empty': '',
                                'none': None,
                            }
                        },
                    ]
                },
            }
        }


class TestSentryMonitorUtils:
    @mock.patch.dict('flask.current_app.config', {
        'SENTRY_ENVIRONMENT': 'testenv'
    })
    @mock.patch(
        'keg_elements.sentry.sentry_sdk.crons.api.capture_checkin', autospec=True, spec_set=True
    )
    def test_job_normal(self, m_sentry_checkin):
        m_sentry_checkin.return_value = 'foo123'

        with sentry.sentry_monitor_job('monitor-key', do_ping=True) as monitor:
            pass

        assert m_sentry_checkin.call_count == 2
        first_call = m_sentry_checkin.call_args_list[0].kwargs
        assert first_call == dict(monitor_slug='monitor-key-testenv', status='in_progress')
        second_call = m_sentry_checkin.call_args_list[1].kwargs
        assert second_call['monitor_slug'] == 'monitor-key-testenv'
        assert second_call['status'] == 'ok'
        assert second_call['check_in_id'] == 'foo123'
        assert second_call['duration']
        assert monitor.checkin_id == 'foo123'

    @mock.patch.dict('flask.current_app.config', {
        'SENTRY_ENVIRONMENT': 'testenv'
    })
    @mock.patch(
        'keg_elements.sentry.sentry_sdk.crons.api.capture_checkin', autospec=True, spec_set=True
    )
    def test_job_sentry_exception(self, m_sentry_checkin):
        m_sentry_checkin.side_effect = Exception('something wrong'), 'foo123'
        with pytest.raises(Exception, match='something wrong'):
            with sentry.sentry_monitor_job('monitor-key', do_ping=True):
                pass

        assert m_sentry_checkin.call_count == 2
        first_call = m_sentry_checkin.call_args_list[0].kwargs
        assert first_call == dict(monitor_slug='monitor-key-testenv', status='in_progress')
        second_call = m_sentry_checkin.call_args_list[1].kwargs
        assert second_call['monitor_slug'] == 'monitor-key-testenv'
        assert second_call['status'] == 'error'

    @mock.patch.dict('flask.current_app.config', {
        'SENTRY_ENVIRONMENT': 'testenv'
    })
    @mock.patch(
        'keg_elements.sentry.sentry_sdk.crons.api.capture_checkin', autospec=True, spec_set=True
    )
    def test_job_override_final_ping(self, m_sentry_checkin):
        m_sentry_checkin.return_value = 'foo123'

        with sentry.sentry_monitor_job('monitor-key', do_ping=True) as monitor:
            monitor.ping_error()

        assert m_sentry_checkin.call_count == 2
        first_call = m_sentry_checkin.call_args_list[0].kwargs
        assert first_call == dict(monitor_slug='monitor-key-testenv', status='in_progress')
        second_call = m_sentry_checkin.call_args_list[1].kwargs
        assert second_call['monitor_slug'] == 'monitor-key-testenv'
        assert second_call['status'] == 'error'
        assert second_call['check_in_id'] == 'foo123'
        assert second_call['duration']

    @mock.patch.dict('flask.current_app.config', {
        'SENTRY_ENVIRONMENT': 'testenv'
    })
    @mock.patch(
        'keg_elements.sentry.sentry_sdk.crons.api.capture_checkin', autospec=True, spec_set=True
    )
    def test_job_no_ping(self, m_sentry_checkin):
        with sentry.sentry_monitor_job('monitor-key', do_ping=False):
            pass

        assert m_sentry_checkin.call_count == 0

    @mock.patch.dict('flask.current_app.config', {
        'SENTRY_ENVIRONMENT': 'testenv'
    })
    @mock.patch(
        'keg_elements.sentry.sentry_sdk.crons.api.capture_checkin', autospec=True, spec_set=True
    )
    def test_job_exception(self, m_sentry_checkin):
        m_sentry_checkin.return_value = 'foo123'

        with pytest.raises(Exception):
            with sentry.sentry_monitor_job('monitor-key', do_ping=True):
                raise ValueError('foo')

        assert m_sentry_checkin.call_count == 2
        first_call = m_sentry_checkin.call_args_list[0].kwargs
        assert first_call == dict(monitor_slug='monitor-key-testenv', status='in_progress')
        second_call = m_sentry_checkin.call_args_list[1].kwargs
        assert second_call['monitor_slug'] == 'monitor-key-testenv'
        assert second_call['status'] == 'error'
        assert second_call['check_in_id'] == 'foo123'
        assert second_call['duration']

    @mock.patch.dict('flask.current_app.config', {
        'SENTRY_ENVIRONMENT': 'testenv'
    })
    @mock.patch(
        'keg_elements.sentry.sentry_sdk.crons.api.capture_checkin', autospec=True, spec_set=True
    )
    def test_config_applied(self, m_sentry_checkin):
        config_data = {
            'jobs': {
                'monitor-key': {
                    'key': 'value',
                }
            }
        }
        sentry.SentryMonitor.apply_config(config_data)
        m_sentry_checkin.assert_called_once_with(
            monitor_slug='monitor-key-testenv',
            status='ok',
            monitor_config={'key': 'value'},
        )

    @mock.patch.dict('flask.current_app.config', {
        'SENTRY_ENVIRONMENT': 'testenv'
    })
    @mock.patch(
        'keg_elements.sentry.sentry_sdk.crons.api.capture_checkin', autospec=True, spec_set=True
    )
    def test_config_error(self, m_sentry_checkin):
        m_sentry_checkin.return_value = None

        config_data = {
            'jobs': {
                'monitor-key': {
                    'key': 'value',
                }
            }
        }
        with pytest.raises(sentry.SentryMonitorConfigError):
            sentry.SentryMonitor.apply_config(config_data)



from keg_elements.forms import Form
from keg_elements.forms.state_field import StateField

from .test_form import FormBase


class StateForm(Form):
    short_long = StateField()
    short_short = StateField(short_labels=True)
    long_long = StateField(short_values=False)
    long_short = StateField(short_values=False, short_labels=True)


class TestStateField(FormBase):
    form_cls = StateForm

    def test_state_choices(self):
        form = self.assert_invalid(short_long="ZZ")
        assert form.errors == {"short_long": ["Not a valid choice"]}

        form = self.assert_valid(short_long="MO")
        form = self.assert_invalid(short_long="Missouri")
        assert form.errors == {"short_long": ["Not a valid choice"]}
        assert ('MO', 'Missouri') in form.short_long.choices
        assert ('Missouri', 'MO') not in form.short_long.choices
        assert ('MO', 'MO') not in form.short_long.choices
        assert ('Missouri', 'Missouri') not in form.short_long.choices

        form = self.assert_valid(long_long="Missouri")
        form = self.assert_invalid(long_long="MO")
        assert form.errors == {"long_long": ["Not a valid choice"]}
        assert ('MO', 'Missouri') not in form.long_long.choices
        assert ('Missouri', 'MO') not in form.long_long.choices
        assert ('MO', 'MO') not in form.long_long.choices
        assert ('Missouri', 'Missouri') in form.long_long.choices

        form = self.assert_valid(short_short="MO")
        form = self.assert_invalid(short_short="Missouri")
        assert form.errors == {"short_short": ["Not a valid choice"]}
        assert ('MO', 'Missouri') not in form.short_short.choices
        assert ('Missouri', 'MO') not in form.short_short.choices
        assert ('MO', 'MO') in form.short_short.choices
        assert ('Missouri', 'Missouri') not in form.short_short.choices

        form = self.assert_valid(long_short="Missouri")
        form = self.assert_invalid(long_short="MO")
        assert form.errors == {"long_short": ["Not a valid choice"]}
        assert ('MO', 'Missouri') not in form.long_short.choices
        assert ('Missouri', 'MO') in form.long_short.choices
        assert ('MO', 'MO') not in form.long_short.choices
        assert ('Missouri', 'Missouri') not in form.long_short.choices



from unittest import mock

import pytest
import sqlalchemy as sa
from keg.db import db

import kegel_app.model.entities as ents
import pytz

from keg_elements.db import columns


class TestColumns(object):

    def test_testing_create_for_timezone_columns(self):
        obj = ents.ColumnTester.fake()
        assert obj.time_zone in pytz.common_timezones

        obj.time_zone = 'other'
        assert obj.time_zone == 'other'
        assert obj.__mapper__.columns.time_zone.type.length == 255

    def test_encrypted_null(self):
        ents.ColumnTester.delete_cascaded()
        obj = ents.ColumnTester.fake(encrypted1=None, encrypted2=None, encrypted3=None)
        assert obj.encrypted1 is None
        assert obj.encrypted2 is None
        assert obj.encrypted3 is None
        results = db.session.execute(sa.text('select * from column_tester')).fetchall()
        assert results[0].encrypted1 is None
        assert results[0].encrypted2 is None
        assert results[0].encrypted3 is None

    def test_store_encrypted(self):
        ents.ColumnTester.delete_cascaded()
        obj = ents.ColumnTester.fake(encrypted1='Foo', encrypted2='Bar', encrypted3='Baz')
        assert obj.encrypted1 == 'Foo'
        assert obj.encrypted2 == 'Bar'
        assert obj.encrypted3 == 'Baz'
        results = db.session.execute(sa.text('select * from column_tester')).fetchall()
        assert 'Foo' not in results[0].encrypted1
        assert 'Bar' not in results[0].encrypted2
        assert results[0].encrypted3 == 'Onm'

    def test_load_encrypted(self):
        ents.ColumnTester.delete_cascaded()
        db.session.execute(sa.text('''
        insert into column_tester (encrypted1, encrypted2, encrypted3) values
            ('gAAAAABYVFHtgRTQJyzkIF5bQ0TiLA3Dm0kfDF5pvSLrZUBrpZiq_f3g3-jUUj8vPmTw2Uf3SW3u73fJNO9ER46xpW9_41IdaQ==',
            'gAAAAABYVFHtbibu6P-oroQ7UYj7g5oiZPqsbeAmxNAIz3wiEjc236bP7c2Ubsf1-S-ANt2r2WNQytxIfwxoE9VyNQmtihCCwA==',
            'Onm'
            )
        '''))
        obj = ents.ColumnTester.query.one()
        assert obj.encrypted1 == 'Foo'
        assert obj.encrypted2 == 'Bar'
        assert obj.encrypted3 == 'Baz'


class TestDBEnum:
    class Status(columns.DBEnum):
        active = 'Active'
        inactive = 'Inactive'
        other = 'Other'

        @classmethod
        def db_name(cls):
            return 'enum_status'

    def test_db_type(self):
        type_ = self.Status.db_type()
        assert type_.name == 'enum_status'
        assert type_.enums == ['active', 'inactive', 'other']

    def test_db_name_required(self):
        class Unnamed(columns.DBEnum):
            pass

        with pytest.raises(NotImplementedError):
            Unnamed.db_name()

    def test_option_pairs(self):
        assert self.Status.option_pairs() == [
            ('active', 'Active'),
            ('inactive', 'Inactive'),
            ('other', 'Other'),
        ]

    def test_form_options(self):
        assert self.Status.form_options() == [
            (self.Status.active, 'Active'),
            (self.Status.inactive, 'Inactive'),
            (self.Status.other, 'Other'),
        ]

    def test_coerce(self):
        assert self.Status.coerce(None) is None

        assert self.Status.coerce(self.Status.active) == self.Status.active

        assert self.Status.coerce('inactive') == self.Status.inactive

        with pytest.raises(ValueError) as exc:
            self.Status.coerce('foo')
        assert str(exc.value) == 'Not a valid selection'

    def test_str(self):
        assert str(self.Status.active) == 'active'
        assert str(self.Status.inactive) == 'inactive'

        assert self.Status.active.__json__() == 'active'
        assert self.Status.inactive.__json__() == 'inactive'

    @mock.patch('keg_elements.db.columns.random', autospec=True, spec_set=True)
    def test_random(self, m_random):
        m_random.choice.side_effect = lambda option: option[1]
        assert self.Status.random() == self.Status.inactive

        m_random.choice.side_effect = lambda option: option[2]
        assert self.Status.random() == self.Status.other



import base64
import contextlib
import re
import sys
import urllib.parse

import flask
import sentry_sdk
from sentry_sdk.integrations.logging import LoggingIntegration as SentryLogging
from sentry_sdk.integrations.flask import FlaskIntegration as SentryFlask

try:
    import requests
except ImportError:
    requests = None

try:
    import yaml
except ImportError:
    yaml = None


class SentryEventFilter:
    """Filter used with Sentry SDK to prevent secrets from being uploaded in exception reports.

    Filtering may happen based on:
    - variable name (e.g. password, key, token)
    - module (e.g. cryptography, itsdangerous)
    - config keys (e.g. SECRET_KEY)

    If you subclass or override, and you are using the install_sentry method to set up Sentry
    SDK, you must pass your event filter instance to install_sentry.

    Types, variable names, and module sets can be provided in the constructor to extend filtering
    on the defaults.
    """

    # Filter out any variables having a type in the following tuple
    sanitized_types = (flask.config.Config,)

    # Filter out values anywhere in the event JSON with a key containing any of the following
    # strings or regular expressions.
    sanitized_var_names = [
        'token',
        'password',
        'secret',
        'passwd',
        'authorization',
        'sentry_dsn',
        'xsrf',
        re.compile('.*key$'),
    ]

    # Filter out all local variables from the stacktrace contained within any of the modules listed.
    # This includes all submodules. Values may be strings or compiled regular expressions.
    sanitized_modules = [
        'keg_elements.crypto',
        'bcrypt',
        'cryptography',
        'encryption',
        'itsdangerous',
        'passlib',
        'PyNaCl',
        'secretstorage',
    ]

    # Filter out all instances of the values stored in the config dict under these keys.
    sanitized_config_keys = [
        'SECRET_KEY',
    ]

    def __init__(
            self,
            types=frozenset(),
            var_names=frozenset(),
            modules=frozenset()
    ):
        self.sanitized_types = tuple(set(self.sanitized_types) | set(types))
        self.sanitized_var_names = set(self.sanitized_var_names) | set(var_names)
        self.sanitized_modules = set(self.sanitized_modules) | set(modules)

        self._variable_regexes = {self._var_to_re(key) for key in self.sanitized_var_names}
        self._module_regexes = {self._module_to_re(mod) for mod in self.sanitized_modules}
        self._config_values = self._config_value_representations()

    def _var_to_re(self, key):
        if isinstance(key, str):
            return re.compile(r'.*{}.*'.format(re.escape(key)), re.IGNORECASE)
        return key

    def _module_to_re(self, module):
        if isinstance(module, str):
            return re.compile(r'^{}(?:\.|$).*'.format(re.escape(module)))
        return module

    def _config_value_representations(self):
        if not flask.current_app:
            return []

        reps = []
        for key in self.sanitized_config_keys:
            base_value = flask.current_app.config.get(key)
            if not base_value:
                continue

            if not isinstance(base_value, (str, bytes)):
                base_value = str(base_value)

            if isinstance(base_value, str):
                bytes_value = base_value.encode(errors='ignore')
                str_value = base_value
            else:
                bytes_value = base_value
                str_value = base_value.decode(errors='ignore')

            reps.extend([
                str_value,
                repr(base_value),
                urllib.parse.quote(bytes_value),
                urllib.parse.quote_plus(bytes_value),
                base64.b64encode(bytes_value).decode(),
                base64.b64encode(bytes_value).rstrip(b'=').decode(),
                base64.b16encode(bytes_value).decode(),
            ])

        return reps

    def repr_process(self, obj, hints):
        if isinstance(obj, self.sanitized_types):
            return f'<Filtered {obj.__class__.__name__}>'
        return NotImplemented

    def should_exclude_var(self, k):
        for key in self._variable_regexes:
            if key.match(k):
                return True
        return False

    def _filter_repr(self, v):
        return f'<Filtered {v.__class__.__name__}>'

    def _filter_value(self, key, value):
        if self.should_exclude_var(key):
            return self._filter_repr(value)
        if isinstance(value, str):
            for conf_val in self._config_values:
                value = value.replace(conf_val, self._filter_repr(conf_val))
        return self._filter_recur(value)

    def _filter_recur(self, obj):
        if isinstance(obj, dict):
            return {
                k: self._filter_value(k, v)
                for k, v in obj.items()
            }
        if isinstance(obj, list):
            return [self._filter_recur(v) for v in obj]
        return obj

    def should_exclude_module(self, m):
        for mod in self._module_regexes:
            if mod.match(m):
                return True
        return False

    def _get_exceptions(self, event):
        try:
            exception = event['exception']
        except KeyError:
            return None

        if isinstance(exception, dict) and 'values' in exception:
            return exception['values']

        return [exception]

    def _filter_modules(self, event):
        exceptions = self._get_exceptions(event)
        if exceptions is None:
            # Event does not appear to include any exception stack traces
            return event

        stacks = []
        for exc in exceptions:
            try:
                stacks.append(exc['stacktrace']['frames'])
            except KeyError:
                # If we cannot process the stack trace, clear everything to minimize the chances of
                # leaking sensitive data in case of an unrecognized event schema
                exc['stacktrace'] = {'frames': []}

        for stack in stacks:
            sanitize_subcalls = False
            for frame in stack:
                module = frame.get('module')
                if sanitize_subcalls or module is None or self.should_exclude_module(module):
                    frame['vars'] = {}
                    # Sanitize any stack frames resulting from calls after entering a sanitized
                    # module. Since the sanitized module may process sensitive data using stdlib
                    # functions or functions from other external libraries.
                    sanitize_subcalls = True
        return event

    def _filter_request(self, event):
        try:
            request_dict = event['request']
        except KeyError:
            return event

        request_dict['cookies'] = {
            k: self._filter_repr(v) for k, v in request_dict.get('cookies', {}).items()
        }
        request_dict['headers'] = {
            k: (self._filter_repr(v) if 'cookie' in k.lower() else v)
            for k, v in request_dict.get('headers', {}).items()
        }

        return event

    def before_send(self, event, hints):
        event = self._filter_recur(event)
        event = self._filter_modules(event)
        event = self._filter_request(event)

        return event


def install_sentry(app, integrations, release=None, event_filter=None, **kwargs):
    """Init Sentry SDK with helpful defaults and an event filter."""
    sentry_dsn = app.config.get('SENTRY_DSN')
    if sentry_dsn is None:
        return

    event_filter = event_filter or SentryEventFilter()
    sentry_sdk.serializer.add_global_repr_processor(event_filter.repr_process)

    sentry_sdk.init(
        dsn=sentry_dsn,
        integrations=[
            SentryLogging(event_level=None),
            SentryFlask(),
            *integrations
        ],
        before_send=event_filter.before_send,
        release=release,
        environment=app.config.get('SENTRY_ENVIRONMENT'),
        include_local_variables=True,
        send_default_pii=True,  # include user details
        **kwargs
    )


class SentryMonitorError(Exception):
    pass


class SentryMonitorConfigError(Exception):
    pass


class SentryMonitor:
    """Sentry allows monitoring scheduled job execution, and exposes this feature in its
    API. SentryMonitor wraps usage.

    Init the class with the "base" monitor key, and the environment to
    tie the monitor to in Sentry.

    Note: environment is not yet well-supported. At the time of this writing, there is no
    support for removing a single environment. So, monitors will currently be single-env,
    and each monitor key is given a suffix of the env name.

    Config YAML format::

        jobs:
            my-cron-job:
                checkin_margin: 1 # in minutes
                schedule:
                  type: crontab
                  value: '30 2 * * *'
                max_runtime: 1 # in minutes
                failure_issue_threshold: 2 # optional
                timezone: Etc/UTC
            other-stats-run:
                checkin_margin: 1
                schedule:
                  type: interval
                  value: 6
                  unit: minute
                max_runtime: 1
                timezone: Etc/UTC


    """
    def __init__(self, key: str, env: str):
        if requests is None:
            raise Exception('requests library required for Sentry monitoring')

        self.key = key
        self.env = env
        self.dsn = flask.current_app.config.get('SENTRY_DSN')
        self.checkin_id = None
        self.start_timestamp = None
        self.status = None

    @classmethod
    def read_config(cls, path):
        """Safe-load YAML from the file at ``path``."""
        if yaml is None:
            raise Exception('pyyaml library required for Sentry monitoring')
        with open(path, 'rb') as f:
            return yaml.safe_load(f)

    @classmethod
    def apply_config(cls, config_data: dict):
        """Runs monitor config on Sentry API for each job in config data."""
        env = flask.current_app.config.get('SENTRY_ENVIRONMENT')
        return_data = {}
        error_data = {}
        for key, monitor_config in config_data.get('jobs', {}).items():
            # monitor gets created via the check-in, which requires status
            monitor = SentryMonitor(key, env)
            try:
                return_data[key] = monitor.configure(monitor_config)
                if not return_data[key]:
                    error_data[key] = return_data[key]
            except SentryMonitorError as exc:
                error_data[key] = str(exc)

        if error_data:
            raise SentryMonitorConfigError(error_data)

        return return_data

    @property
    def monitor_key(self):
        return f'{self.key}-{self.env}'.lower()

    def configure(self, payload):
        return sentry_sdk.crons.api.capture_checkin(
            monitor_slug=self.monitor_key,
            status=sentry_sdk.crons.consts.MonitorStatus.OK,
            monitor_config=payload,
        )

    def get_duration(self):
        duration = None
        if self.start_timestamp:
            duration = sentry_sdk.utils.now() - self.start_timestamp
        return duration

    def ping_status(self, status, duration=None):
        self.status = status
        kwargs = {}
        if duration is not None:
            kwargs['duration'] = duration
        if self.checkin_id is not None:
            kwargs['check_in_id'] = self.checkin_id
        self.checkin_id = sentry_sdk.crons.api.capture_checkin(
            monitor_slug=self.monitor_key, status=status, **kwargs
        )
        return self.checkin_id

    def ping_ok(self):
        return self.ping_status(
            sentry_sdk.crons.consts.MonitorStatus.OK,
            duration=self.get_duration(),
        )

    def ping_error(self):
        return self.ping_status(
            sentry_sdk.crons.consts.MonitorStatus.ERROR,
            duration=self.get_duration(),
        )

    def ping_in_progress(self):
        self.start_timestamp = sentry_sdk.utils.now()
        return self.ping_status(sentry_sdk.crons.consts.MonitorStatus.IN_PROGRESS)


@contextlib.contextmanager
def sentry_monitor_job(key: str, env: str = None, do_ping: bool = False):
    """Context manager for running in_progress, then ok status pings on a Sentry monitor.

    Sentry will not be pinged unless the ``do_ping`` kwarg is True (default is False).

    ``env`` will default to ``SENTRY_ENVIRONMENT`` config value if not specified.

    Context manager returns the ``SentryMonitor`` instance. The wrapped code may ping a
    different ending status (such as if an error result should occur without an
    exception). Exceptions will result in an error status."""
    env = env or flask.current_app.config.get('SENTRY_ENVIRONMENT')
    if not do_ping:
        yield
    else:
        monitor = SentryMonitor(key, env)

        try:
            monitor.ping_in_progress()

            yield monitor

            if monitor.status == sentry_sdk.crons.consts.MonitorStatus.IN_PROGRESS:
                # Code wrapped in context likely has not manually pinged another status.
                # This block allows us to manually set an error based on app logic.
                monitor.ping_ok()
        except Exception:
            monitor.ping_error()
            exc_info = sys.exc_info()
            sentry_sdk.utils.reraise(*exc_info)



from unittest import mock

import flask
import flask_webtest
from pyquery import PyQuery
from webgrid.extensions import RequestArgsLoader

from kegel_app.model import entities as ents


class TestDemoForm:
    def test_get(self):
        client = flask_webtest.TestApp(flask.current_app)
        resp = client.get('/demo-form')
        assert 'Input field' in resp

    def test_post_valid(self):
        client = flask_webtest.TestApp(flask.current_app)
        resp = client.get('/demo-form')
        resp.form['my_field'] = 'foo'
        resp = resp.form.submit()
        assert resp.flashes == [('success', 'Form submitted successfully.')]
        assert resp.location.endswith('demo-grid')

    def test_post_invalid(self):
        client = flask_webtest.TestApp(flask.current_app)
        resp = client.get('/demo-form')
        resp = resp.form.submit()
        assert resp.flashes == [('error', 'Form errors detected, see below for details.')]
        assert resp.status_code == 200
        assert 'Input field' in resp


class TestDemoGrid:
    def setup_method(self):
        ents.Thing.delete_cascaded()

    def test_get(self):
        ents.Thing.fake(name='Thing Foo')
        ents.Thing.fake(name='Thing Two')
        client = flask_webtest.TestApp(flask.current_app)
        resp = client.get('/demo-grid')
        assert 'datagrid' in resp
        assert 'Thing Foo' in resp
        assert 'Thing Two' in resp
        assert PyQuery(resp.body)('form.header').attr('method') == 'post'

    def test_qs_args_applied(self):
        ents.Thing.fake(name='Thing Foo')
        ents.Thing.fake(name='Thing Two')
        client = flask_webtest.TestApp(flask.current_app)
        resp = client.get('/demo-grid?op(name)=contains&v1(name)=Foo')
        assert 'datagrid' in resp
        assert 'Thing Foo' in resp
        assert 'Thing Two' not in resp

    def test_post(self):
        ents.Thing.fake(name='Thing Foo')
        ents.Thing.fake(name='Thing Two')
        client = flask_webtest.TestApp(flask.current_app)
        resp = client.post('/demo-grid?session_key=abc123', {
            'op(name)': 'contains',
            'v1(name)': 'Foo',
        }, status=302)
        assert resp.location.endswith('/demo-grid?session_key=abc123')
        resp = resp.follow()
        assert 'datagrid' in resp
        assert 'Thing Foo' in resp
        assert 'Thing Two' not in resp

    @mock.patch('kegel_app.grids.DemoGrid.session_on', False)
    def test_post_no_session(self):
        ents.Thing.fake(name='Thing Foo')
        ents.Thing.fake(name='Thing Two')
        client = flask_webtest.TestApp(flask.current_app)
        resp = client.post('/demo-grid?session_key=abc123', {
            'op(name)': 'contains',
            'v1(name)': 'Foo',
        })
        assert 'datagrid' in resp
        assert 'Thing Foo' in resp
        assert 'Thing Two' not in resp

    @mock.patch('kegel_app.grids.DemoGrid.manager.args_loaders', [RequestArgsLoader])
    def test_post_no_form_loader(self):
        client = flask_webtest.TestApp(flask.current_app)
        client.post('/demo-grid?session_key=abc123', {
            'op(name)': 'contains',
            'v1(name)': 'Foo',
        }, status=405)

    def test_export(self):
        ents.Thing.fake(name='Thing Foo')
        client = flask_webtest.TestApp(flask.current_app)
        resp = client.get('/demo-grid?export_to=xlsx')
        assert 'spreadsheetml' in resp.content_type



import codecs
import datetime
import enum
import random
import uuid

import sqlalchemy as sa
from keg.db import db

import keg_elements.db.columns as columns
import keg_elements.db.mixins as mixins

many_things_mapper = db.Table(
    'many_things_mapper',
    sa.Column('thing_id', sa.Integer, sa.ForeignKey('things.id', ondelete='cascade')),
    sa.Column('many_id', sa.Integer, sa.ForeignKey('many_things.id', ondelete='cascade'))
)


class Units(enum.Enum):
    feet = 'ft'
    inches = 'in'
    meters = 'm'


class ThingStatus(columns.DBEnum):
    open = 'Open'
    closed = 'Closed'

    @classmethod
    def db_name(cls):
        return 'enum_thing_status'


def random_scale_check():
    return 12.3456


class Thing(mixins.DefaultMixin, db.Model):
    __tablename__ = 'things'

    name = db.Column(db.Unicode(50), nullable=False)
    color = db.Column(db.Unicode, info=dict(randomdata='random_color'))
    scale_check = db.Column(db.Numeric(8, 4), info=dict(randomdata=random_scale_check))
    float_check = db.Column(db.Float)
    float_check_prop = sa.orm.column_property(float_check + 6)
    units = db.Column(sa.Enum(Units, name='enum_units'))
    status = db.Column(ThingStatus.db_type())
    date_check = db.Column(db.Date, default=datetime.date.today, onupdate=datetime.date.today)

    @sa.ext.hybrid.hybrid_property
    def name_and_color(self):
        return '{}-{}'.format(self.name, self.color)

    @name_and_color.expression
    def name_and_color(cls):
        return cls.name + sa.sql.literal('-') + cls.color

    @sa.ext.hybrid.hybrid_property
    def hybrid_name(self):
        return self.name

    @hybrid_name.expression
    def hybrid_name(cls):
        return cls.name

    @sa.ext.hybrid.hybrid_property
    def hybrid_id(self):
        return self.id

    @hybrid_id.expression
    def hybrid_id(cls):
        return cls.id

    @classmethod
    def random_color(cls):
        return 'blue'

    @classmethod
    def delete_cascaded(cls):
        RelatedThing.delete_cascaded()
        AncillaryA.delete_cascaded()
        AncillaryB.delete_cascaded()
        super().delete_cascaded()


class RelatedThing(mixins.DefaultMixin, db.Model):
    __tablename__ = 'related_things'

    name = db.Column(db.Unicode(50), nullable=False)
    is_enabled = db.Column(db.Boolean, nullable=False, default=False,
                           server_default=sa.false())

    thing_id = sa.Column(sa.Integer, sa.ForeignKey(Thing.id), nullable=False)
    thing = sa.orm.relationship(
        lambda: Thing,
        backref=sa.orm.backref(
            "related_things",
            cascade="all, delete-orphan",
            cascade_backrefs=False,
        ),
        cascade_backrefs=False,
    )

    @classmethod
    def fake(cls, **kwargs):
        cls.testing_set_related(kwargs, Thing)
        return super(RelatedThing, cls).fake(**kwargs)


class OtherThing(mixins.DefaultMixin, db.Model):
    __tablename__ = 'other_things'

    name = db.Column(db.Unicode(50), nullable=False)
    is_enabled = db.Column(db.Boolean, nullable=False, default=False,
                           server_default=sa.false())
    unique_field = db.Column(db.Unicode(50), unique=True, nullable=False)

    thing_id = sa.Column(sa.Integer, sa.ForeignKey(Thing.id, ondelete='cascade'), nullable=False)
    thing = sa.orm.relationship(
        lambda: Thing,
        backref=sa.orm.backref(
            "other_things",
            cascade="all, delete-orphan",
            cascade_backrefs=False,
        ),
        cascade_backrefs=False,
    )

    @classmethod
    def fake(cls, **kwargs):
        cls.testing_set_related(kwargs, Thing)
        kwargs.setdefault('unique_field', str(uuid.uuid4()))
        return super(OtherThing, cls).fake(**kwargs)


class ManyToManyThing(mixins.DefaultMixin, db.Model):
    __tablename__ = 'many_things'

    name = db.Column(db.Unicode(50), nullable=False)

    things = sa.orm.relationship(
        lambda: Thing,
        secondary=many_things_mapper,
        backref=sa.orm.backref("manys", cascade_backrefs=False),
    )


class ThingWithRequiredBoolean(mixins.DefaultMixin, db.Model):
    __tablename__ = 'required_boolean_table'

    nullable_boolean = db.Column(db.Boolean, nullable=True)
    required_boolean = db.Column(db.Boolean, nullable=False)
    required_boolean_with_default = db.Column(db.Boolean, nullable=False, default=False)
    required_boolean_with_server_default = db.Column(db.Boolean, nullable=False,
                                                     server_default='false')


class MultiplePrimaryKeys(mixins.DefaultMixin, db.Model):
    __tablename__ = 'multikey_table'

    other_pk = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.Unicode(50))


def super_secure_encrypt(data, key):
    return codecs.encode(data, 'rot_13').encode()


def super_secure_decrypt(data, key):
    return codecs.encode(data.decode(), 'rot_13')


class ColumnTester(mixins.DefaultMixin, db.Model):
    __tablename__ = 'column_tester'

    time_zone = db.Column(columns.TimeZoneType(length=100))
    encrypted1 = db.Column(columns.EncryptedUnicode(key=lambda: b'a' * 32))
    encrypted2 = db.Column(columns.EncryptedUnicode(key=b'b' * 32))
    encrypted3 = db.Column(columns.EncryptedUnicode(
        key=b'c' * 32,
        encrypt=super_secure_encrypt,
        decrypt=super_secure_decrypt
    ))


class AncillaryA(mixins.DefaultMixin, db.Model):
    __tablename__ = 'ancillary_as'

    thing_id = sa.Column(sa.Integer, sa.ForeignKey(Thing.id), nullable=False)

    __table_args__ = (
        sa.UniqueConstraint('id', thing_id),
    )

    @classmethod
    def fake(cls, **kwargs):
        if 'thing_id' not in kwargs:
            kwargs['thing_id'] = Thing.fake().id

        return super(AncillaryA, cls).fake(**kwargs)

    @classmethod
    def delete_cascaded(cls):
        UsesBoth.delete_cascaded()
        super().delete_cascaded()


class AncillaryB(mixins.DefaultMixin, db.Model):
    __tablename__ = 'ancillary_bs'

    thing_id = sa.Column(sa.Integer, sa.ForeignKey(Thing.id), nullable=False)

    __table_args__ = (
        sa.UniqueConstraint('id', thing_id),
    )

    @classmethod
    def fake(cls, **kwargs):
        if 'thing_id' not in kwargs:
            kwargs['thing_id'] = Thing.fake().id

        return super(AncillaryB, cls).fake(**kwargs)

    @classmethod
    def delete_cascaded(cls):
        UsesBoth.delete_cascaded()
        super().delete_cascaded()


# Foreign keys to non-unique, non-primary key columns are not supported by MSSQL
class UsesBoth(mixins.DefaultMixin, db.Model):
    __tablename__ = 'uses_boths'
    __table_args__ = (
        sa.ForeignKeyConstraint(
            ('ancillary_a_id', 'thing_id'),
            (AncillaryA.id, AncillaryA.thing_id)
        ),
        sa.ForeignKeyConstraint(
            ('ancillary_b_id', 'thing_id'),
            (AncillaryB.id, AncillaryB.thing_id)
        ),
    )

    thing_id = sa.Column(sa.Integer, nullable=False)
    ancillary_a_id = sa.Column(sa.Integer, nullable=False)
    ancillary_b_id = sa.Column(sa.Integer, nullable=False)

    @classmethod
    def fake(cls, **kwargs):
        if 'thing_id' not in kwargs:
            kwargs['thing_id'] = Thing.fake().id

        thing_id = kwargs['thing_id']

        if 'ancillary_a_id' not in kwargs:
            kwargs['ancillary_a_id'] = AncillaryA.fake(thing_id=thing_id).id

        if 'ancillary_b_id' not in kwargs:
            kwargs['ancillary_b_id'] = AncillaryB.fake(thing_id=thing_id).id

        return super(UsesBoth, cls).fake(**kwargs)


class ConstraintTester(mixins.DefaultMixin, db.Model):
    unique1 = sa.Column(sa.Integer, nullable=False, unique=True)
    unique2 = sa.Column(sa.Integer, nullable=False)

    check = sa.Column(sa.Integer, nullable=False)

    __tablename__ = 'constraint_tester'
    __table_args__ = (
        sa.UniqueConstraint(unique2, name='uq_constraint_tester_unique2'),
        sa.CheckConstraint(check <= 100, name='ck_constraint_tester_check'),
    )

    @classmethod
    def fake(cls, **kwargs):
        kwargs.setdefault('check', random.randint(0, 100))
        return super(ConstraintTester, cls).fake(**kwargs)


class HardDeleteParent(mixins.DefaultMixin, db.Model):
    __tablename__ = 'hard_delete_parent'


class SoftDeleteTester(mixins.SoftDeleteMixin, mixins.DefaultMixin, db.Model):
    __tablename__ = 'softdelete_tester'

    hdp_id = sa.Column(
        sa.Integer,
        sa.ForeignKey(HardDeleteParent.id),
        nullable=False
    )
    hdp = sa.orm.relationship(
        HardDeleteParent,
        backref=sa.orm.backref(
            "sdts", cascade="all,delete-orphan", cascade_backrefs=False
        ),
        cascade_backrefs=False,
    )

    @classmethod
    def fake(cls, **kwargs):
        kwargs['hdp_id'] = kwargs.get('hpd_id') or HardDeleteParent.fake().id
        return super().fake(**kwargs)


class DefaultNumeric(mixins.DefaultMixin, db.Model):
    __tablename__ = 'default_numeric'

    number = sa.Column(sa.Numeric)


class LookupTester(mixins.LookupMixin, mixins.DefaultMixin, db.Model):
    pass



from keg.app import Keg
import keg.db

from kegel_app.extensions import Grid
from kegel_app.views import keg_element_blueprint


class KegElApp(Keg):
    import_name = 'kegel_app'
    keyring_enable = False
    use_blueprints = [keg_element_blueprint]
    db_enabled = True

    def on_init_complete(self):
        Grid.manager.init_db(keg.db.db)
        Grid.manager.init_app(self)



import sys

from keg.db import db
import pytest
from pyquery import PyQuery as pq
from werkzeug.datastructures import MultiDict
import wtforms as wtf
from wtforms import validators
import wtforms_components

import keg_elements.forms as ke_forms
import kegel_app.model.entities as ents
from keg_elements.forms import (
    FieldMeta,
    Form,
    ModelForm,
    RelationshipField,
    RelationshipMultipleField,
    SelectField,
)


class FormBase(object):
    form_cls = None
    entity_cls = None

    def compose_meta(self, fields_meta_cls=None, **data):
        assert self.entity_cls is not None, 'must set entity_cls on {}'\
            .format(self.__class__.__name__)

        class _ComposedForm(ModelForm):
            class Meta:
                model = self.entity_cls
                csrf = False

            if fields_meta_cls:
                FieldsMeta = fields_meta_cls
            else:
                class FieldsMeta:
                    __default__ = FieldMeta
        return _ComposedForm(MultiDict(data))

    def create_form(self, **data):
        form_cls = data.pop('form_cls', self.form_cls)
        return form_cls(MultiDict(data))

    def assert_invalid(self, **data):
        form = self.create_form(**data)
        assert not form.validate()
        return form

    def assert_valid(self, **data):
        form = self.create_form(**data)
        assert form.validate(), form.errors
        return form


class FruitForm(Form):
    fruit = SelectField(validators=[validators.Optional()],
                        choices=[('apple', 'apple'), ('banana', 'banana')])
    letter = SelectField(validators=[validators.Optional()],
                         choices=[('a', 'a'), ('b', 'b'), ('', 'blank')], add_blank_choice=False)
    letter2 = SelectField(validators=[validators.InputRequired()], choices=[('a', 'a'), ('b', 'b')])
    number = SelectField(validators=[validators.Optional()], choices=[(1, '1'), (0, '0')])
    numstr = SelectField(validators=[validators.Optional()], choices=[('1', '1'), ('0', '0')],
                         coerce=str)


class TestSelectField(FormBase):
    form_cls = FruitForm

    def test_blank_choice_rendering(self):
        form = FruitForm(meta={'csrf': False})
        fruit_options = pq(form.fruit())('option')
        assert fruit_options.length == 3
        assert fruit_options.eq(0).attr('value') == ''
        assert fruit_options.eq(1).attr('value') == 'apple'

        letter_options = pq(form.letter())('option')
        assert letter_options.length == 3

    def test_blank_choice_submit(self):
        form = self.assert_valid(**{'fruit': '', 'letter': '', 'letter2': 'a'})

        assert form.fruit.data is None
        assert form.letter.data == ''

    def test_blank_choice_required(self):
        form = self.assert_invalid(**{'letter2': ''})
        assert form.letter2.errors == ['This field is required.']

        form = self.assert_valid(**{'letter2': 'b'})

    def test_blank_choice_coerce(self):
        self.assert_valid(**{'letter2': 'a', 'number': '1'})

        # make sure we can override the int helper by passing through an explicit coerce
        self.assert_valid(**{'letter2': 'a', 'numstr': '1'})

    def test_db_enum_default_choices(self):
        class ThingForm(ModelForm):
            class Meta:
                model = ents.Thing
                csrf = False

            class FieldsMeta:
                __default__ = FieldMeta

        form = ThingForm()
        assert form.status.choices == [
            (ents.ThingStatus.open, 'Open'),
            (ents.ThingStatus.closed, 'Closed'),
        ]
        assert form.status.coerce('open') is ents.ThingStatus.open

    def test_db_enum_override_choices(self):
        class ThingForm(ModelForm):
            class Meta:
                model = ents.Thing
                csrf = False

            class FieldsMeta:
                __default__ = FieldMeta
                status = FieldMeta(choices=[('1', 'Foo'), ('2', 'Bar')], coerce=str)

        form = ThingForm()
        assert form.status.choices == [
            ('1', 'Foo'),
            ('2', 'Bar'),
        ]
        assert form.status.coerce(1) == '1'


class TestRequiredBoolRadioField(FormBase):
    class RequiredBoolMockForm(Form):
        is_competent = ke_forms.RequiredBoolRadioField()
        statement_is = ke_forms.RequiredBoolRadioField(true_label='True', false_label='False')

    form_cls = RequiredBoolMockForm

    def test_required_bool_radio_rendering(self):
        form = self.RequiredBoolMockForm(meta={'csrf': False})
        is_competent_labels = pq(form.is_competent())('label')
        assert is_competent_labels.length == 2
        assert is_competent_labels.eq(0).text() == 'Yes'
        assert is_competent_labels.eq(1).text() == 'No'

        statement_is_labels = pq(form.statement_is())('label')
        assert statement_is_labels.length == 2
        assert statement_is_labels.eq(0).text() == 'True'
        assert statement_is_labels.eq(1).text() == 'False'

    def test_required_bool_radio_is_required(self):
        form = self.assert_invalid()
        assert form.is_competent.errors == ['This field is required.']
        assert form.statement_is.errors == ['This field is required.']

        form = self.assert_invalid(**{'is_competent': 'True'})
        assert form.is_competent.errors == []
        assert form.statement_is.errors == ['This field is required.']

        form = self.assert_invalid(**{'statement_is': 'True'})
        assert form.is_competent.errors == ['This field is required.']
        assert form.statement_is.errors == []

        form = self.assert_valid(**{'is_competent': 'False', 'statement_is': 'False'})


class TestDefaultTypeOfRequiredBooleanField(FormBase):
    entity_cls = ents.ThingWithRequiredBoolean

    def test_field_types(self):
        form = self.compose_meta()
        assert type(form.nullable_boolean) is wtf.fields.BooleanField
        assert type(form.required_boolean) is ke_forms.RequiredBoolRadioField
        assert type(form.required_boolean_with_default) is wtf.fields.BooleanField
        assert type(form.required_boolean_with_server_default) is ke_forms.RequiredBoolRadioField


class TestFieldMeta(FormBase):
    entity_cls = ents.Thing

    def test_required_unspecified(self):
        form = self.compose_meta()
        assert form.name.flags.required
        assert not form.color.flags.required

    def test_required_true(self):
        class Temp:
            __default__ = FieldMeta(required=True)
        form = self.compose_meta(Temp)
        assert form.name.flags.required
        assert form.color.flags.required

    def test_required_false(self):
        class Temp:
            __default__ = FieldMeta(required=False)
        form = self.compose_meta(Temp)
        assert not form.name.flags.required
        assert not form.color.flags.required

    def test_widget_no_override(self):
        form = self.compose_meta()
        assert type(form.color.widget) is wtforms_components.widgets.TextInput

    def test_widget_override(self):
        class WidgetOverrideFieldsMeta:
            __default__ = FieldMeta
            color = FieldMeta(widget=wtf.widgets.TextArea())

        form = self.compose_meta(fields_meta_cls=WidgetOverrideFieldsMeta)
        assert type(form.color.widget) is wtf.widgets.TextArea

    def test_extra_validators(self):
        class ExtraValidatorsFieldsMeta:
            def _is_roy(form, field):
                if field.data not in {'red', 'orange', 'yellow'}:
                    raise wtf.validators.ValidationError('Not a ROY color')

            __default__ = FieldMeta
            color = FieldMeta(extra_validators=[_is_roy])

        form = self.compose_meta(fields_meta_cls=ExtraValidatorsFieldsMeta,
                                 name='Test', color='red')
        assert form.validate()

        form = self.compose_meta(fields_meta_cls=ExtraValidatorsFieldsMeta,
                                 name='Test', color='mauve')
        assert not form.validate()
        assert set(form.color.errors) == {'Not a ROY color'}

    def test_coerce(self):
        class CoerceFieldsMeta:
            __default__ = FieldMeta
            units = FieldMeta(coerce=lambda x: ents.Units[x.lower()] if x else x,
                              choices=[(x, x.value) for x in ents.Units])

        form = self.compose_meta(fields_meta_cls=CoerceFieldsMeta, name='Test', units='FEET')
        assert isinstance(form.units, wtf.SelectField)
        assert form.validate()
        assert isinstance(form.units.data, ents.Units)
        assert form.units.data == ents.Units.feet

        # coerce not allowed for non-select fields
        class BadCoerceFieldsMeta:
            __default__ = FieldMeta
            name = FieldMeta(coerce=lambda x: x)
        with pytest.raises(ValueError) as exc:
            self.compose_meta(fields_meta_cls=BadCoerceFieldsMeta)
        assert str(exc.value) == '`coerce` argument may only be used for select fields'

    def test_default(self):
        class DefaultFieldsMeta:
            __default__ = FieldMeta
            name = FieldMeta(default='foo')

        form = self.compose_meta(fields_meta_cls=DefaultFieldsMeta)
        assert form.name.default == 'foo'
        assert form.color.default is None


class TestValidators(FormBase):
    entity_cls = ents.Thing

    def test_numeric_scale_check(self):
        # too many places
        form = self.compose_meta(scale_check='0.11111')
        form.validate()
        assert form.scale_check.errors == ['Field must have no more than 4 decimal places.']

        # equal to data type
        form = self.compose_meta(scale_check='0.1111')
        form.validate()
        assert form.scale_check.errors == []

        # fewer places than data type
        form = self.compose_meta(scale_check='0.111')
        form.validate()
        assert form.scale_check.errors == []

        # integer
        form = self.compose_meta(scale_check='5')
        form.validate()
        assert form.scale_check.errors == []

        # test that other validation happens when type is wrong
        form = self.compose_meta(scale_check='aaa')
        form.validate()
        assert 'Not a valid decimal value.' in form.scale_check.errors

    def test_numeric_range_check(self):
        message = 'Number must be between -9999.9999 and 9999.9999.'

        # too high
        form = self.compose_meta(scale_check='10000.0000')
        form.validate()
        assert form.scale_check.errors == [message]

        # too low
        form = self.compose_meta(scale_check='-10000.0000')
        form.validate()
        assert form.scale_check.errors == [message]

        # top limit
        form = self.compose_meta(scale_check='9999.9999')
        form.validate()
        assert form.scale_check.errors == []

        # bottom limit
        form = self.compose_meta(scale_check='-9999.9999')
        form.validate()
        assert form.scale_check.errors == []

    def test_validators_not_added_for_float(self):
        form = self.compose_meta(float_check=sys.float_info.max)
        form.validate()
        assert form.float_check.errors == []

        form = self.compose_meta(float_check=sys.float_info.min)
        form.validate()
        assert form.float_check.errors == []

        assert len(form.float_check.validators) == 1

    def test_numeric_scale_precision_required(self):
        with pytest.raises(ValueError, match='Numeric fields must specify precision and scale'):
            class TestForm1(ke_forms.ModelForm):
                class Meta:
                    model = ents.DefaultNumeric

        class TestForm2(ke_forms.ModelForm):
            class Meta:
                model = ents.DefaultNumeric
                exclude = ('number',)

            number = wtf.DecimalField('Number', validators=[])

    def test_length_validation_not_applied_for_enums(self):
        form = self.compose_meta()
        for validator in form.units.validators:
            assert not isinstance(validator, wtf.validators.Length)


class FeaturesForm(Form):
    _form_ident_enabled = False
    name = wtf.StringField(validators=[validators.data_required()])
    color = wtf.StringField()


class NumbersSubForm(wtf.Form):
    number = wtf.StringField('Number', validators=[validators.data_required()])
    color = wtf.StringField('Color')


class NumbersForm(Form):
    _form_ident_enabled = False
    numbers = wtf.FieldList(wtf.FormField(NumbersSubForm), min_entries=2)
    numbers2 = wtf.FieldList(wtf.StringField('Number'), min_entries=2)


class TestFieldsToDict(FormBase):

    def test_simple(self):
        self.form_cls = FeaturesForm
        form = self.assert_invalid(color='blue')
        expected = {
            'name': {
                'data': None,
                'errors': ['This field is required.'],
                'label': 'Name',
                'required': True,
            },
            'color': {
                'data': 'blue',
                'errors': [],
                'label': 'Color',
                'required': None
            },
        }
        assert form.fields_todict() == expected

    def test_simple_json(self):
        self.form_cls = FeaturesForm
        form = self.assert_invalid(color='blue')
        expected = {
            'name': {
                'data': None,
                'errors': ['This field is required.'],
                'label': 'Name',
                'required': True,
            },
            'color': {
                'data': 'blue',
                'errors': [],
                'label': 'Color',
                'required': None
            },
        }
        fields_data = form.fields_todict()
        assert fields_data == expected

    def test_field_to_dict_field(self):
        form = NumbersSubForm()
        form.validate()
        assert ke_forms.field_to_dict(form.number) == {
            'data': None,
            'errors': ['This field is required.'],
            'label': 'Number',
            'required': True
        }

    def test_field_to_dict_select(self):
        form = FruitForm(letter='b')
        form.validate()
        assert ke_forms.field_to_dict(form.letter) == \
            {'data': 'b', 'errors': [], 'label': 'Letter', 'required': None}
        assert ke_forms.field_to_dict(form.letter2) == {
            'data': None,
            'errors': ['This field is required.'],
            'label': 'Letter2',
            'required': True
        }

    def test_field_to_dict_fieldlist(self):
        form = NumbersForm()
        form.validate()
        expected = [
            {
                'data': None,
                'errors': [],
                'label': 'Number',
                'required': None,
            },
            {
                'data': None,
                'errors': [],
                'label': 'Number',
                'required': None,
            }
        ]
        assert ke_forms.field_to_dict(form.numbers2) == expected

    def test_field_to_dict_form_fieldlist(self):
        data = {
            'numbers-0-number': '123',
            'numbers-1-color': 'blue',
        }
        form = NumbersForm(MultiDict(data))
        form.validate()
        expected = [
            {
                'number': {
                    'data': '123',
                    'errors': [],
                    'label': 'Number',
                    'required': True
                },
                'color': {
                    'data': None,
                    'errors': [],
                    'label': 'Color',
                    'required': None
                }
            },
            {
                'number': {
                    'data': None,
                    'errors': ['This field is required.'],
                    'label': 'Number',
                    'required': True
                },
                'color': {
                    'data': 'blue',
                    'errors': [],
                    'label': 'Color',
                    'required': None
                }
            },
        ]
        assert ke_forms.field_to_dict(form.numbers) == expected

    def test_nested(self):
        self.form_cls = NumbersForm
        data = {
            'numbers-0-number': '123',
            'numbers-1-color': 'blue'
        }
        form = self.assert_invalid(**data)
        expected = {
            'numbers': [
                {
                    'number': {
                        'data': '123',
                        'errors': [],
                        'label': 'Number',
                        'required': True,
                    },
                    'color': {
                        'data': None,
                        'errors': [],
                        'label': 'Color',
                        'required': None,
                    }
                },
                {
                    'number': {
                        'data': None,
                        'errors': ['This field is required.'],
                        'label': 'Number',
                        'required': True,
                    },
                    'color': {
                        'data': 'blue',
                        'errors': [],
                        'label': 'Color',
                        'required': None,
                    }
                },
            ],
            'numbers2': [
                {
                    'data': None,
                    'errors': [],
                    'label': 'Number',
                    'required': None,
                },
                {
                    'data': None,
                    'errors': [],
                    'label': 'Number',
                    'required': None
                }
            ]
        }
        fields_dict = form.fields_todict()
        assert fields_dict == expected, fields_dict


class TestFieldOrder():
    def test_field_missing_in_order(self):
        class MissingFields(Form):
            _field_order = ('num3', 'num2', 'num4')
            num1 = wtf.IntegerField()
            num2 = wtf.IntegerField()
            num3 = wtf.IntegerField()

        form = MissingFields()

        with pytest.raises(ValueError) as e:
            list(form)

        assert 'num1' in str(e.value)
        assert 'num4' in str(e.value)

    def test_with_csrf(self):
        # This whole hullabaloo is to get CSRF enabled for the form since we normally turn it off
        # for testing.
        class CSRFImpl(wtf.csrf.core.CSRF):
            def generate_csrf_token(self, token):
                return 'token'

        class CSRF(Form):
            _form_ident_enabled = False
            _field_order = ('num2', 'num1',)

            class Meta:
                csrf = True
                csrf_class = CSRFImpl
                csrf_secret = '123'

            num1 = wtf.IntegerField()
            num2 = wtf.IntegerField()

        form = CSRF()
        assert [x.name for x in form] == ['csrf_token', 'num2', 'num1']

    def test_field_order(self):
        class OrderedForm(Form):
            _field_order = ('num3', 'num1', 'num2',)
            num1 = wtf.IntegerField()
            num2 = wtf.IntegerField()
            num3 = wtf.IntegerField()

        form = OrderedForm()

        assert [x.name for x in form] == ['keg_form_ident', 'num3', 'num1', 'num2']

    def test_field_order_no_ident(self):
        class OrderedForm(Form):
            _form_ident_enabled = False
            _field_order = ('num3', 'num1', 'num2',)
            num1 = wtf.IntegerField()
            num2 = wtf.IntegerField()
            num3 = wtf.IntegerField()

        form = OrderedForm()

        assert [x.name for x in form] == ['num3', 'num1', 'num2']

    def test_field_unorder(self):
        class UnorderedForm(Form):
            num1 = wtf.IntegerField()
            num2 = wtf.IntegerField()
            num3 = wtf.IntegerField()

        form = UnorderedForm()

        assert [x.name for x in form] == ['num1', 'num2', 'num3', 'keg_form_ident']


class TestFormIdentValidation(FormBase):
    class MyForm(Form):
        num1 = wtf.IntegerField()

    form_cls = MyForm

    def test_form_ident_validated(self):
        form = self.assert_invalid(keg_form_ident='foo')
        assert form.form_errors == []
        assert form.errors == {'keg_form_ident': ['Invalid value, must be one of: my-form.']}
        self.assert_valid(keg_form_ident='my-form')

    def test_form_ident_validated_custom_key(self):
        class TestForm(self.MyForm):
            @classmethod
            def _form_ident_key(cls):
                return 'mycoolfield'

        form = self.assert_invalid(form_cls=TestForm, mycoolfield='foo')
        assert form.form_errors == []
        assert form.errors == {'mycoolfield': ['Invalid value, must be one of: test-form.']}
        self.assert_valid(form_cls=TestForm, mycoolfield='test-form')

    def test_form_ident_validated_custom_key_invalid(self):
        with pytest.raises(Exception, match='Cannot start form ident name with "_"'):
            class TestForm(self.MyForm):
                @classmethod
                def _form_ident_key(cls):
                    return '_mycoolfield'

    def test_form_ident_validated_custom_value(self):
        class TestForm(self.MyForm):
            @classmethod
            def _form_ident_value(cls):
                return 'bar'

        form = self.assert_invalid(form_cls=TestForm, keg_form_ident='foo')
        assert form.form_errors == []
        assert form.errors == {'keg_form_ident': ['Invalid value, must be one of: bar.']}
        self.assert_valid(form_cls=TestForm, keg_form_ident='bar')

    def test_form_ident_not_strict(self):
        class TestForm(self.MyForm):
            _form_ident_strict = False

        self.assert_valid(form_cls=TestForm, keg_form_ident='foo')

    def test_form_ident_not_exists(self):
        class TestForm(self.MyForm):
            _form_ident_enabled = False

        form = TestForm()
        assert not getattr(TestForm, 'keg_form_ident', None)
        assert not getattr(form, 'keg_form_ident', None)
        assert 'keg_form_ident' not in form._fields


class TestFormLevelValidation(FormBase):
    class MyForm(Form):
        num1 = wtf.IntegerField()
        num2 = wtf.IntegerField()
        num3 = wtf.IntegerField()

        @ke_forms.form_validator
        def equal_42(self):
            if self.num1.data + self.num2.data != 42:
                raise wtf.ValidationError('Does not add up')

        @ke_forms.form_validator
        def in_order(self):
            if not (self.num1.data <= self.num2.data <= self.num3.data):
                raise wtf.ValidationError('Out of order')

    form_cls = MyForm

    def test_form_valid(self):
        form = self.assert_valid(num1=5, num2=37, num3=100)
        assert form.form_errors == []
        assert form.errors == {}

    def test_form_invalid(self):
        form = self.assert_invalid(num1=40, num2=3, num3=50)
        assert form.form_errors == ['Does not add up', 'Out of order']
        assert form.errors == {None: ['Does not add up', 'Out of order']}

    def test_stop_validation_with_error(self):
        class StopValidationForm(Form):
            s1 = wtf.StringField()
            s2 = wtf.StringField()

            @ke_forms.form_validator
            def fields_equal(self):
                if self.s1.data != self.s2.data:
                    raise wtf.validators.StopValidation('not equal')

            @ke_forms.form_validator
            def other_validator(self):
                assert False, 'Validation should have stopped'  # pragma: no cover

        form = self.assert_invalid(form_cls=StopValidationForm, s1='v1', s2='v2')
        assert form.form_errors == ['not equal']
        assert form.errors == {None: ['not equal']}

    def test_stop_validation_no_error(self):
        class StopValidationForm(Form):
            s1 = wtf.StringField()
            s2 = wtf.StringField()

            @ke_forms.form_validator
            def fields_equal(self):
                if self.s1.data != self.s2.data:
                    raise wtf.validators.StopValidation

            @ke_forms.form_validator
            def other_validator(self):
                assert False, 'Validation should have stopped'  # pragma: no cover

        form = self.assert_valid(form_cls=StopValidationForm, s1='v1', s2='v2')
        assert form.form_errors == []
        assert form.errors == {}

    def test_invalid_with_field_errors(self):
        class InvalidFieldsForm(Form):
            s1 = wtf.StringField(validators=[wtf.validators.Length(max=3)])
            s2 = wtf.StringField()

            @ke_forms.form_validator
            def fields_equal(self):
                if self.s1.data != self.s2.data:
                    raise wtf.ValidationError('not equal')

        form = self.assert_invalid(form_cls=InvalidFieldsForm, s1='1234', s2='4321')
        assert form.form_errors == ['not equal']
        assert form.errors == {
            None: ['not equal'],
            's1': ['Field cannot be longer than 3 characters.']
        }

    def test_do_not_validate_with_field_errors(self):
        class InvalidFieldsForm(Form):
            s1 = wtf.StringField(validators=[wtf.validators.Length(max=3)])
            s2 = wtf.StringField()

            @ke_forms.form_validator(only_when_fields_valid=True)
            def my_validator(self):
                assert False, 'Validation should not run'  # pragma: no cover

        form = self.assert_invalid(form_cls=InvalidFieldsForm, s1='1234', s2='4321')
        assert form.form_errors == []
        assert form.errors == {'s1': ['Field cannot be longer than 3 characters.']}

    def test_validators_inherited(self):
        class SubclassForm(self.MyForm):
            @ke_forms.form_validator
            def num3_is_even(self):
                if self.num3.data % 2:
                    raise wtf.ValidationError('Num3 is odd')

            @ke_forms.form_validator
            def equal_42(self):
                if self.num1.data * self.num2.data != 42:
                    raise wtf.ValidationError('Does not compute')

        form = self.assert_invalid(num1=7, num2=5, num3=51, form_cls=SubclassForm)
        assert form.form_errors == ['Out of order', 'Num3 is odd', 'Does not compute']
        assert form.errors == {
            None: ['Out of order', 'Num3 is odd', 'Does not compute']
        }

        form = self.assert_valid(num1=6, num2=7, num3=50, form_cls=SubclassForm)
        assert form.form_errors == []
        assert form.errors == {}


class TestExcludesDatetimes(FormBase):
    entity_cls = ents.Thing

    def test_exclude_default_cols(self):
        form = self.compose_meta()
        assert 'updated_utc' not in form.fields_todict()
        assert 'created_utc' not in form.fields_todict()


class TestTypeHintingTextInputB4:
    class DemoForm(wtf.Form):
        prefix_field = wtf.StringField(widget=ke_forms.TypeHintingTextInputB4(prefix="foo"))
        suffix_field = wtf.StringField(widget=ke_forms.TypeHintingTextInputB4(suffix="bar"))
        both_field = wtf.StringField(widget=ke_forms.TypeHintingTextInputB4(prefix="a", suffix="z"))

    def test_prefix(self):
        form = self.DemoForm()
        rendered = pq(form.prefix_field())

        group = rendered(".input-group")
        assert len(group) == 1

        children = group.children()
        assert len(children) == 2

        prefix, field = children.items()
        assert prefix.has_class("input-group-prepend")
        assert len(prefix.children(".input-group-text"))
        assert prefix.children(".input-group-text").text() == "foo"

        assert field.is_("input")

    def test_suffix(self):
        form = self.DemoForm()
        rendered = pq(form.suffix_field())

        group = rendered(".input-group")
        assert len(group) == 1

        children = group.children()
        assert len(children) == 2

        field, suffix = children.items()
        assert suffix.has_class("input-group-append")
        assert len(suffix.children(".input-group-text"))
        assert suffix.children(".input-group-text").text() == "bar"

        assert field.is_("input")

    def test_both(self):
        form = self.DemoForm()
        rendered = pq(form.both_field())

        group = rendered(".input-group")
        assert len(group) == 1

        children = group.children()
        assert len(children) == 3

        prefix, field, suffix = children.items()

        assert prefix.has_class("input-group-prepend")
        assert len(prefix.children(".input-group-text"))
        assert prefix.children(".input-group-text").text() == "a"

        assert suffix.has_class("input-group-append")
        assert len(suffix.children(".input-group-text"))
        assert suffix.children(".input-group-text").text() == "z"

        assert field.is_("input")

    def test_text_escaped(self):
        class Form(wtf.Form):
            prefix = wtf.StringField(widget=ke_forms.TypeHintingTextInputB4(prefix="<script>"))
            suffix = wtf.StringField(widget=ke_forms.TypeHintingTextInputB4(suffix="&foo&"))

        form = Form()

        prefix = form.prefix()
        assert "&lt;script&gt;" in prefix

        suffix = form.suffix()
        assert "&amp;foo&amp;" in suffix


class TestTypeHintingTextInputB3:
    class DemoForm(wtf.Form):
        prefix_field = wtf.StringField(widget=ke_forms.TypeHintingTextInputB3(prefix="foo"))
        suffix_field = wtf.StringField(widget=ke_forms.TypeHintingTextInputB3(suffix="bar"))
        both_field = wtf.StringField(widget=ke_forms.TypeHintingTextInputB3(prefix="a", suffix="z"))

    def test_prefix(self):
        form = self.DemoForm()
        rendered = pq(form.prefix_field())

        group = rendered(".input-group")
        assert len(group) == 1

        children = group.children()
        assert len(children) == 2

        prefix, field = children.items()
        assert prefix.has_class("input-group-addon")
        assert prefix.text() == "foo"

        assert field.is_("input")

    def test_suffix(self):
        form = self.DemoForm()
        rendered = pq(form.suffix_field())

        group = rendered(".input-group")
        assert len(group) == 1

        children = group.children()
        assert len(children) == 2

        field, suffix = children.items()
        assert suffix.has_class("input-group-addon")
        assert suffix.text() == "bar"

        assert field.is_("input")

    def test_both(self):
        form = self.DemoForm()
        rendered = pq(form.both_field())

        group = rendered(".input-group")
        assert len(group) == 1

        children = group.children()
        assert len(children) == 3

        prefix, field, suffix = children.items()

        assert prefix.has_class("input-group-addon")
        assert prefix.text() == "a"

        assert suffix.has_class("input-group-addon")
        assert suffix.text() == "z"

        assert field.is_("input")

    def test_text_escaped(self):
        class Form(wtf.Form):
            prefix = wtf.StringField(widget=ke_forms.TypeHintingTextInputB3(prefix="<script>"))
            suffix = wtf.StringField(widget=ke_forms.TypeHintingTextInputB3(suffix="&foo&"))

        form = Form()

        prefix = form.prefix()
        assert "&lt;script&gt;" in prefix

        suffix = form.suffix()
        assert "&amp;foo&amp;" in suffix


class TestModelForm:
    class MyForm(ModelForm):
        class Meta:
            model = ents.LookupTester

        class FieldsMeta:
            pass

    def test_default_title_case(self):
        form = TestModelForm.MyForm(None)
        assert form.label.label.text == 'Label'


def test_fields_meta_inheritance():
    class FormA(ModelForm):
        class FieldsMeta:
            a = 'foo'

    class FormB(FormA):
        class FieldsMeta:
            b = 'bar'

    form = FormB()
    assert form.FieldsMeta.a == 'foo'
    assert form.FieldsMeta.b == 'bar'


class RelationshipMixin:
    def assert_object_options(self, field, option_objects, has_blank=True):
        object_options = [(str(obj.id), obj.name) for obj in option_objects]
        expected_choices = [
            ('', ''),
            *object_options
        ] if has_blank else object_options
        index = 0
        for value, label, selected, *other_args in field.iter_choices():
            assert expected_choices[index] == (value, label)
            index += 1


class ThingForeignKeyRelationshipMixin:
    def setup_method(self):
        ents.Thing.delete_cascaded()
        db.session.remove()

    def get_field(self, form):
        raise NotImplementedError()  # pragma: no cover

    def create_formdata(self, form):
        raise NotImplementedError()  # pragma: no cover

    def coerce(self, value):
        return value

    def test_relationship_options(self):
        ents.Thing.fake(name='Foo')
        thing1 = ents.Thing.fake()
        form = self.create_form(
            self.create_relationship(lambda this: this.orm_cls.name != 'Foo')
        )

        self.assert_object_options(self.get_field(form), [thing1])

    def test_options_include_form_obj_value(self):
        foo_thing = ents.Thing.fake(name='Foo')
        thing1 = ents.Thing.fake(name='Thing 1')
        related_thing = ents.RelatedThing.fake(thing=foo_thing, name='Related')
        form = self.create_form(
            self.create_relationship(lambda this: this.orm_cls.name != 'Foo'),
            obj=related_thing,
        )

        self.assert_object_options(self.get_field(form), [foo_thing, thing1])

    def test_options_sorted(self):
        thing_b = ents.Thing.fake(name='BBB')
        thing_c = ents.Thing.fake(name='CCC')
        thing_a = ents.Thing.fake(name='AAA')
        form = self.create_form(
            self.create_relationship(lambda this: this.orm_cls.name != 'Foo')
        )

        self.assert_object_options(self.get_field(form), [thing_a, thing_b, thing_c])

    def test_no_query_filter(self):
        foo_thing = ents.Thing.fake(name='Foo')
        thing1 = ents.Thing.fake(name='Thing 1')
        form = self.create_form(self.create_relationship())

        self.assert_object_options(self.get_field(form), [foo_thing, thing1])


class TestForeignKeyRelationship(ThingForeignKeyRelationshipMixin, RelationshipMixin):
    """Test a foreign key field."""
    def create_relationship(self, query_filter=None):
        return RelationshipField('Thing', ents.Thing, 'name', query_filter=query_filter)

    def create_form(self, relationship, **kwargs):
        class RelatedThingForm(ModelForm):
            thing_id = relationship

            class Meta:
                model = ents.RelatedThing

        return RelatedThingForm(**kwargs)

    def get_field(self, form):
        return form.thing_id

    def test_coerce_formdata(self):
        thing = ents.Thing.fake()
        form = self.create_form(self.create_relationship(),
                                formdata=MultiDict({'thing_id': str(thing.id)}))
        assert form.thing_id.data == thing.id


class TestOrmRelationship(ThingForeignKeyRelationshipMixin, RelationshipMixin):
    """Test a relationship field that corresponds to a SA RelationshipProperty."""

    def create_relationship(self, query_filter=None):
        return RelationshipField('Thing', ents.Thing, 'name', query_filter=query_filter,
                                 fk_attr=None)

    def create_form(self, relationship, **kwargs):
        class RelatedThingForm(ModelForm):
            thing = relationship

            class Meta:
                model = ents.RelatedThing

        return RelatedThingForm(**kwargs)

    def coerce(self, value):
        return ents.Thing.get(value)

    def get_field(self, form):
        return form.thing

    def test_coerce_formdata(self):
        thing = ents.Thing.fake()
        form = self.create_form(self.create_relationship(),
                                formdata=MultiDict({'thing': str(thing.id)}))
        assert form.thing.data == thing


class TestOrmRelationshipOrmAttr(TestOrmRelationship):
    """Test a relationship field that uses an ORM attr for the label."""

    def create_relationship(self, query_filter=None):
        return RelationshipField('Thing', ents.Thing, ents.Thing.name, query_filter=query_filter,
                                 fk_attr=None)


class TestOrmRelationshipOrmFkAttr(TestOrmRelationship):
    """Test a relationship field that uses an ORM attr for the fk."""

    def create_relationship(self, query_filter=None):
        return RelationshipField('Thing', ents.Thing, 'name', query_filter=query_filter,
                                 fk_attr=ents.Thing.id)

    def create_form(self, relationship, **kwargs):
        class RelatedThingForm(ModelForm):
            thing_id = relationship

            class Meta:
                model = ents.RelatedThing

        return RelatedThingForm(**kwargs)

    def get_field(self, form):
        return form.thing_id

    def test_coerce_formdata(self):
        thing = ents.Thing.fake()
        form = self.create_form(self.create_relationship(),
                                formdata=MultiDict({'thing_id': str(thing.id)}))
        assert form.thing_id.data == thing.id


class TestOrmRelationshipOrmHybridAttr(TestOrmRelationshipOrmAttr):
    """Test a relationship field that uses an ORM hybrid property attr for the label."""

    def create_relationship(self, query_filter=None):
        return RelationshipField('Thing', ents.Thing, ents.Thing.hybrid_name,
                                 query_filter=query_filter, fk_attr=None)


class TestOrmRelationshipOrmHybridFkAttr(TestOrmRelationshipOrmFkAttr):
    """Test a relationship field that uses an ORM hybrid property for the fk."""

    def create_relationship(self, query_filter=None):
        return RelationshipField('Thing', ents.Thing, 'name', query_filter=query_filter,
                                 fk_attr=ents.Thing.hybrid_id)


class TestRelationshipFieldGenerator:
    def setup_method(self):
        ents.Thing.delete_cascaded()
        db.session.expunge_all()

    def create_form(self, **kwargs):
        class RelatedThingForm(ModelForm):
            class Meta:
                model = ents.RelatedThing
                include_foreign_keys = True

        return RelatedThingForm(**kwargs)

    def create_custom_form(self, field_kwargs, **kwargs):
        class RelatedThingForm(ModelForm):
            class Meta:
                model = ents.RelatedThing

            thing_id = RelationshipField('Thing', ents.Thing, **field_kwargs)

        return RelatedThingForm(**kwargs)

    def test_field_created(self):
        form = self.create_form()
        assert form.thing_id
        assert form.thing_id.label.text == 'Thing'

    def test_choices(self):
        thing1 = ents.Thing.fake(name='foo')
        thing2 = ents.Thing.fake(name='bar')
        form = self.create_form()
        assert form.thing_id.choice_values == ['', thing2.id, thing1.id]

    def test_validators_applied(self):
        form = self.create_form()
        assert isinstance(form.thing_id.validators[0], validators.InputRequired)

    def test_custom_label_attr(self):
        thing1 = ents.Thing.fake(name='foo', color='blue')
        thing2 = ents.Thing.fake(name='bar', color='red')
        form = self.create_custom_form({'label_attr': 'color'})
        assert form.thing_id.choice_values == ['', thing1.id, thing2.id]

    def test_choices_filtered(self):
        ents.Thing.fake(name='foo', color='blue')
        thing2 = ents.Thing.fake(name='bar', color='red')
        form = self.create_custom_form({'query_filter': ents.Thing.name != 'foo'})
        assert form.thing_id.choice_values == ['', thing2.id]

    def test_choices_included(self):
        thing1 = ents.Thing.fake(name='foo', color='blue')
        thing2 = ents.Thing.fake(name='bar', color='red')
        related_thing = ents.RelatedThing.fake(thing=thing1)
        form = self.create_custom_form({'query_filter': ents.Thing.name != 'foo'},
                                       obj=related_thing)
        assert form.thing_id.choice_values == ['', thing2.id, thing1.id]

    def test_choices_data_filter_not_applied_incorrectly(self):
        thing1 = ents.Thing.fake(name='foo', color='blue')
        thing2 = ents.Thing.fake(name='bar', color='red')
        thing3 = ents.Thing.fake(name='baz', color='purple')
        related_thing = ents.RelatedThing.fake(thing=thing1)
        form = self.create_custom_form({'query_filter': None}, obj=related_thing)
        assert form.thing_id.choice_values == ['', thing2.id, thing3.id, thing1.id]


class TestCollectionRelationship(RelationshipMixin):
    """Test field that corresponds to a SA collection."""
    def create_relationship(self, query_filter=None):
        return RelationshipMultipleField('Related Things', ents.RelatedThing, 'name',
                                         query_filter=query_filter)

    def assert_object_options(self, field, option_objects):
        return super().assert_object_options(field, option_objects, has_blank=False)

    def create_form(self, relationship, **kwargs):
        class ThingForm(ModelForm):
            related_things = relationship

            class Meta:
                model = ents.Thing

        return ThingForm(**kwargs)

    def setup_method(self):
        ents.Thing.delete_cascaded()

    def test_relationship_options(self):
        ents.RelatedThing.fake(name='Foo')
        thing1 = ents.RelatedThing.fake()
        form = self.create_form(
            self.create_relationship(lambda this: this.orm_cls.name != 'Foo')
        )

        self.assert_object_options(form.related_things, [thing1])

    def test_options_include_form_obj_value(self):
        foo_thing = ents.RelatedThing.fake(name='Foo')
        thing1 = ents.RelatedThing.fake(name='Thing 1')
        thing2 = ents.RelatedThing.fake(name='Thing 2')
        thing_obj = ents.Thing.fake(related_things=[thing1, thing2])
        form = self.create_form(
            self.create_relationship(lambda this: ~this.orm_cls.id.in_([thing1.id, thing2.id])),
        )
        self.assert_object_options(form.related_things, [foo_thing])

        form = self.create_form(
            self.create_relationship(lambda this: ~this.orm_cls.id.in_([thing1.id, thing2.id])),
            obj=thing_obj,
        )
        self.assert_object_options(form.related_things, [foo_thing, thing1, thing2])

    def test_load_save_data(self):
        class ManyThingsForm(ModelForm):
            things = RelationshipMultipleField('Things', ents.Thing)

            class Meta:
                model = ents.ManyToManyThing

        thing1 = ents.Thing.fake()
        thing2 = ents.Thing.fake()
        thing3 = ents.Thing.fake()
        manything = ents.ManyToManyThing.fake(things=[thing1, thing2])

        form = ManyThingsForm(obj=manything)
        # form data level should be ORM instances
        assert form.things.data == [thing1, thing2]

        form = ManyThingsForm(
            obj=manything,
            formdata=MultiDict([
                ('things', str(thing2.id)),
                ('things', str(thing3.id)),
            ])
        )
        assert form.things.data == [thing2, thing3]
        form.populate_obj(manything)
        assert manything.things == [thing2, thing3]



import contextlib
import math
import random
from decimal import Decimal

import sqlalchemy as sa
from sqlalchemy.sql import expression
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.types import DateTime
import sqlalchemy_utils as sautils

from blazeutils.strings import randchars

from keg.db import db
from keg_elements.extensions import lazy_gettext as _


class utcnow(expression.FunctionElement):
    type = DateTime()


@compiles(utcnow, 'postgresql')
def _pg_utcnow(element, compiler, **kw):
    return "TIMEZONE('utc', CURRENT_TIMESTAMP)"


@compiles(utcnow, 'mssql')
def _ms_utcnow(element, compiler, **kw):
    return "GETUTCDATE()"


@compiles(utcnow, 'sqlite')
def _sqlite_utcnow(element, compiler, **kw):
    return "CURRENT_TIMESTAMP"


def has_column(orm_cls_or_table, column_name):
    """Searches an entity for a column having the given key.

    While a column's default key is the entity attribute name, a column may be given
    an explicit key instead. This method uses sqlalchemy-utils to find the column.
    """
    for column in sautils.get_columns(orm_cls_or_table):
        if column.key == column_name:
            return True

    return False


def range_from_column(column):
    if 'random_range' in column.info:
        rand_range = column.info['random_range']
        if not isinstance(rand_range, tuple) or len(rand_range) != 2:
            raise Exception(
                'random_range info for {} expected to be tuple with low and high'.format(
                    column.key
                ))
        return rand_range

    if 'random_magnitude' in column.info:
        rand_mag = column.info['random_magnitude']
        return (-rand_mag, rand_mag)

    return None


def random_int(column, default_range):
    """Find a random number that satisfies the given column's data size and meta info.

    :param column: SA column to find a range by type/info and generate a random number.
    :param default_range: Tuple. Fallback range if type/info does not provide a range.
    :returns: Random integer.
    """
    column_range = range_from_column(column)
    if column_range:
        return random.randint(*column_range)

    if type(column.type) is sa.Integer:
        exponent = 31
    elif type(column.type) is sa.SmallInteger:
        exponent = 15
    elif type(column.type) is sa.BigInteger:
        exponent = 63
    else:
        return random.randint(*default_range)
    magnitude = (2 ** exponent) - 1
    return random.randint(-magnitude, magnitude)


def random_numeric(column):
    """Find a random number that satisfies the given column's precision and scale."""
    column_range = range_from_column(column)
    if column_range:
        return random.uniform(*column_range)

    fractional_digits = column.type.scale
    whole_digits = column.type.precision - fractional_digits

    # only use about half the digits to make arithmetic done with this less likely to overflow
    max_whole = 10 ** math.ceil(whole_digits / 2.0) - 1

    whole = random.randint(-max_whole, max_whole)

    fractional = Decimal(random.randint(0, 10 ** fractional_digits - 1)) / 10 ** fractional_digits
    return fractional + whole


def validate_unique_exc(exc, constraint_name=None):
    """Check the given exception to see if it indicates a duplicate key error.

    Supports SQLite, PostgreSQL, and MSSQL dialects.

    Optionally provide a constraint_name kwarg for a stricter test.
    """
    return _validate_unique_msg(db.engine.dialect.name, str(exc), constraint_name)


def _validate_unique_msg(dialect, msg, constraint_name=None):
    """
        Does the heavy lifting for validate_unique_exception().

        Broken out separately for easier unit testing.  This function takes string args.
    """
    if constraint_name is not None and dialect != 'sqlite' and constraint_name not in msg:
        return False

    if dialect == 'postgresql':
        return 'duplicate key value violates unique constraint' in msg
    elif dialect == 'mssql':
        return 'Cannot insert duplicate key' in msg
    elif dialect == 'sqlite':
        return 'UNIQUE constraint failed' in msg
    else:
        raise ValueError('is_unique_exc() does not yet support dialect: %s' % dialect)


@contextlib.contextmanager
def raises_unique_exc(constraint_name):
    """pytest helper and context manager to ensure an exception result is a duplicate key error."""
    import pytest
    with pytest.raises(sa.exc.IntegrityError) as exc:
        yield
    assert validate_unique_exc(exc.value, constraint_name)


@contextlib.contextmanager
def raises_check_exc(contraint_name):
    """pytest helper and context manager to ensure an exception result is a check constraint
       error."""
    import pytest
    with pytest.raises(sa.exc.IntegrityError) as exc:
        yield
    assert _is_check_const(db.engine.dialect.name, str(exc.value), contraint_name)


def _is_check_const(dialect, msg, constraint_name):
    if dialect == 'mssql':
        return 'conflicted with the CHECK constraint' in msg and constraint_name in msg
    elif dialect == 'sqlite':
        return (
            'CHECK constraint {} failed'.format(constraint_name) in msg
            or 'CHECK constraint failed' in msg
        )
    elif dialect == 'postgresql':
        return 'violates check constraint' in msg and constraint_name in msg
    else:
        raise ValueError('is_constraint_exc() does not yet support dialect: %s' % dialect)


def randemail(length, randomizer=randchars):
    """Generate a random email address at the given length.
    :param length: must be at least 7 or the function will throw a ValueError.
    :param randomizer: is a function for generating random characters. It must have an identical
    interface to `randchars`. The default function is `randchars`.
    """
    if length < 7:
        raise ValueError(_('length must be at least 7'))

    half = (length - 2 - 3) / 2.0  # 2 characters for @ and . and 3 for TLD
    return (randomizer(int(math.floor(half)), 'alphanumeric')
            + '@' + randomizer(int(math.ceil(half)), 'alphanumeric')
            + '.' + randomizer(3, 'alpha'))


def session_commit():
    """Commit the db session, and roll back if there is a failure.

    Raises the exception that caused the rollback."""
    try:
        db.session.commit()
    except Exception:
        db.session.rollback()
        raise


def session_flush():
    """Flush the db session, and roll back if there is a failure.

    Raises the exception that caused the rollback."""
    try:
        db.session.flush()
    except Exception:
        db.session.rollback()
        raise


class CollectionUpdater(object):
    """Update a collection attribute of a model object.

    Takes a model object, the attribute name of the collection to be updated, and
    the data to update the collection with. If a record in the given data matches
    a record in the collection, the record is edited. Otherwise, the given record is
    added to the collection.

    All records in the collection that are not included in the given data are removed
    from the collection.
    """

    def __init__(self, entity, attr_name, data):
        self.entity = entity
        self.attr_name = attr_name
        self.data = data
        self.keep_children = set()
        ent_cls = entity.__class__

        with db.session.no_autoflush:
            self.collection = getattr(entity, attr_name)
            queryable_attr = getattr(ent_cls, attr_name)
            # Unless they sent in the wrong type, rel_prop should be a
            # sqlalchemy.orm.properties.RelationshipProperty instance
            rel_prop = queryable_attr.property
            # now that we have the property, go through the mapper to get to the child entity class
            self.child_cls = rel_prop.mapper.class_

    def find_child(self, data):
        """Find the child record associated with the related object"""
        with db.session.no_autoflush:
            child_entity_keys = self.child_cls.primary_keys()
            supplied_keys = set(data.get(column.key)
                                for column in child_entity_keys)

            if None in supplied_keys:
                return None

            for child in self.collection:
                record_keys = set(getattr(child, col.key)
                                  for col in child_entity_keys)
                if record_keys == supplied_keys:
                    return child
            return None

    def update(self):
        """Update the objects associated with the entity

        Objects are matched by keys found in supplied data, if any. If no keys are supplied,
        the existing objects in the collection are removed, and new records are constructed
        from the data.

        The way SQLAlchemy handles collection updates presents a limitation for collections
        having unique constraints. If we drop/recreate the record with the same unique value,
        we will get a unique constraint exception. To work around this, we cache info from
        matched objects in state, then remove/flush objects first. This clears the table for
        the new records from adds/edits.
        """
        with db.session.no_autoflush:
            to_add = []
            to_edit = []
            for record in self.data:
                child = self.find_child(record)
                state = sa.inspect(child) if child else None

                if state and state.persistent:
                    to_edit.append((child, record))
                else:
                    to_add.append(record)

                self.keep_children.add(child)

            self._remove_unmodified()
            session_flush()

            for child, record in to_edit:
                child.edit(_commit=False, **record)

            for record in to_add:
                child = self.child_cls.add(_commit=False, **record)
                self.collection.append(child)

    def _remove_unmodified(self):
        remove_children = [child for child in self.collection if child not in self.keep_children]
        for child in remove_children:
            self.collection.remove(child)



from wtforms import fields as wtfields
from wtforms.validators import input_required

from keg_elements.forms import Form


class DemoForm(Form):
    my_field = wtfields.StringField('Input field', validators=[input_required()])
